# Standard Library
import os
import shutil

# Pyramid
from pyramid.config import Configurator

import pytest

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.http.static import DefaultStaticAssetPolicy
from websauna.system.tests.utils import execute_command
from websauna.tests.webserver import customized_web_server


HERE = os.path.dirname(__file__)
STATIC_CONF_FILE = os.path.join(HERE, "static-asset-test.ini")


@pytest.fixture(scope="module")
def cache_app(request):
    """Construct a WSGI app with static asset caching enabled."""
    request = init_websauna(STATIC_CONF_FILE)
    return request.app


@pytest.fixture
def caching_web_server(request, cache_app):
    server = customized_web_server(request, cache_app)
    return server()


def test_collect_static_asset():
    """Collect static files and stash them with MD5 sums."""
    c = Configurator()
    sap = DefaultStaticAssetPolicy(c)

    sap.add_static_view("websauna-static", "websauna.system:static")
    collected = sap.collect_static()

    # Check one resource from the collectin to see we succeeded
    assert collected["websauna-static"]["pyramid-32x32.png"] == 'perma-asset/pyramid-32x32.c453183eee6627ff09e49f0384cededd.png'


def test_collect_recurse():
    """Check another more complicated static file folder collect"""
    c = Configurator()
    sap = DefaultStaticAssetPolicy(c)

    sap.add_static_view("deform-static", "deform:static")
    collected = sap.collect_static()

    assert len(collected) > 0

    # Check one resource from the collectin to see we succeeded
    assert collected["deform-static"]["pickadate/translations/ja_JP.js"].startswith('perma-asset/pickadate/translations/ja_JP.')


def test_map_static_asset(browser, caching_web_server):
    """Use collected information to return static URLs"""

    cache = os.path.join("websauna", "system", "static", "perma-asset")
    if os.path.exists(cache):
        shutil.rmtree(cache)

    # Run static asset collecteor
    execute_command(["ws-collect-static", STATIC_CONF_FILE], folder=os.getcwd(), timeout=30.0)

    b = browser
    b.visit(caching_web_server)

    el = b.find_by_css("link[rel='stylesheet']")[0]._element
    bootstrap_css = el.get_attribute("href")

    assert "perma-asset" in bootstrap_css
    b.visit(bootstrap_css)

    # Identify correct CSS load by having some known text inside CSS payload
    assert "Bootstrap" in b.html



from websauna.system.http import Request

from .tasks import WebsaunaTask


class TaskFinished:
    """This task is fired when a Celery task finishes regardless if the task failed or not.

    Intended to be used to clean up thread current context sensitive data.

    This is called before ``request._process_finished_callbacks()`` is called. This is **not** called when tasks are executed eagerly.
    """

    def __init__(self, request: Request, task: WebsaunaTask):
        self.request = request
        self.task = task



"""Events fired by websauna.system.user package."""
# Websauna
from websauna.system.http import Request
from websauna.system.user.models import User


class UserEvent:
    """User related event."""

    def __init__(self, request: Request, user: User):
        """Initialize UserEvent.

        :param request: Pyramid Request.
        :param user: User object.
        """
        self.request = request
        self.user = user


class UserCreated(UserEvent):
    """User is created.

    Fired upon:

        * Social media login
        * When sending email activation link
    """


class FirstLogin(UserEvent):
    """User logs in to the site for the first time.

    Fired upon:

        * Social media login
        * After clicking email activation link

    Fired before :py:class:`Login`
    """


class Login(UserEvent):
    """User has logged in to the site.

    Fired upon:

        * Social media login
        * When sending email activation link

    Fired after login data (last_login_ip) has been updated.
    """


class UserAuthSensitiveOperation(UserEvent):
    """User authentication details have changes.

    All user sessions should be dropped.

    Fired upon:

        * Password change
        * Email change (TODO)
    """

    def __init__(self, request: Request, user: User, kind: str):
        """Initialize UserAuthSensitiveOperation.

        :param request: Pyramid Request.
        :param user: User object.
        :param kind: What kind of operation triggered this event. Free to application to choose. E.g. "password_change", "email_change", "username_change", "enabled_change".
        """
        super().__init__(request, user)
        self.kind = kind


class NewRegistrationEvent(UserEvent):
    """User register to the portal."""

    def __init__(self, request: Request, user: User, activation: str, values: dict):
        """Initialize NewRegistrationEvent.

        :param request: Pyramid Request.
        :param user: User object.
        :param activation: User activation code.
        :param values: User data.
        """
        super().__init__(request, user)
        self.activation = activation
        self.values = values


class RegistrationActivatedEvent(UserEvent):
    """User activation om the portal."""

    def __init__(self, request: Request, user: User, activation: str):
        """Initialize RegistrationActivatedEvent.

        :param request: Pyramid Request.
        :param user: User object.
        :param activation: User activation code.
        """
        super().__init__(request, user)
        self.activation = activation


class PasswordResetEvent(UserEvent):
    """User reset password."""

    def __init__(self, request: Request, user: User, password: str):
        """Initialize PasswordResetEvent.

        :param request: Pyramid Request.
        :param user: User object.
        :param password: User password.
        """
        super(PasswordResetEvent, self).__init__(request, user)
        self.password = password



"""ws-sync-db script.

Create initial tables for the database specified on the configuration file.
"""
# Standard Library
import sys
import typing as t

# Pyramid
import transaction

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message
from websauna.system.model.meta import Base


def main(argv: t.List[str]=sys.argv):
    """Create initial tables for the database specified on the configuration file.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv)

    config_uri = get_config_uri(argv)
    request = init_websauna(config_uri)

    with transaction.manager:
        engine = request.dbsession.get_bind()
        # Always enable UUID extension for PSQL
        # TODO: Convenience for now, because we assume UUIDs, but make this somehow configurable
        engine.execute('CREATE EXTENSION IF NOT EXISTS "uuid-ossp"')

        Base.metadata.create_all(engine)


if __name__ == "__main__":
    main()



"""Helper functions to initializer Websauna framework for command line applications."""
# Standard Library
import logging
import os
import sys
import typing as t

# Pyramid
import plaster
from pyramid import router
from pyramid import scripting

from rainbow_logging_handler import RainbowLoggingHandler

# Websauna
from websauna.system import Initializer
from websauna.system.http import Request
from websauna.system.http.utils import make_routable_request
from websauna.system.model.meta import create_dbsession


def prepare_config_uri(config_uri: str) -> str:
    """Make sure a configuration uri has the prefix ws://.

    :param config_uri: Configuration uri, i.e.: websauna/conf/development.ini
    :return: Configuration uri with the prefix ws://.
    """
    if not config_uri.startswith('ws://'):
        config_uri = 'ws://{uri}'.format(uri=config_uri)
    return config_uri


def get_wsgi_app(config_uri: str, defaults: dict) -> router.Router:
    """Return a Websauna WSGI application given a configuration uri.

    :param config_uri: Configuration uri, i.e.: websauna/conf/development.ini.
    :param defaults: Extra options to be passed to the app.
    :return: A Websauna WSGI Application
    """
    config_uri = prepare_config_uri(config_uri)
    loader = plaster.get_loader(config_uri)
    return loader.get_wsgi_app(defaults=defaults)


def initializer_from_app(app: router.Router) -> Initializer:
    """Return the initializer for the given app.

    :param app: Websauna WSGI application
    :return: Websauna Initializer
    """
    initializer = getattr(app, 'initializer', None)
    assert initializer is not None, "Configuration did not yield to Websauna application with Initializer set up"
    return initializer


def setup_logging(config_uri, disable_existing_loggers=False):
    """Include-aware Python logging setup from INI config file.
    """
    config_uri = prepare_config_uri(config_uri)
    loader = plaster.get_loader(config_uri, protocols=['wsgi'])
    loader.setup_logging(disable_existing_loggers=disable_existing_loggers)


def setup_console_logging(log_level: t.Optional[str]=None):
    """Setup console logging.

    Aimed to give easy sane defaults for logging in command line applications.

    Don't use logging settings from INI, but use hardcoded defaults.
    """

    formatter = logging.Formatter("[%(asctime)s] [%(name)s %(funcName)s] %(message)s")  # same as default

    # setup `RainbowLoggingHandler`
    # and quiet some logs for the test output
    handler = RainbowLoggingHandler(sys.stdout)
    handler.setFormatter(formatter)
    logger = logging.getLogger()
    logger.handlers = [handler]

    env_level = os.environ.get("LOG_LEVEL", "info")
    log_level = log_level or getattr(logging, env_level.upper())
    logger.setLevel(log_level)

    logger = logging.getLogger("requests.packages.urllib3.connectionpool")
    logger.setLevel(logging.ERROR)

    # SQL Alchemy transactions
    logger = logging.getLogger("txn")
    logger.setLevel(logging.ERROR)


def init_websauna(config_uri: str, sanity_check: bool=False, console_app: bool=False, extra_options: dict=None) -> Request:
    """Initialize Websauna WSGI application for a command line oriented script.

    Example:

    .. code-block:: python

        import sys
        from websauna.system.devop.cmdline import init_websauna

        config_uri = sys.argv[1]
        request = init_websauna(config_uri)

    :param config_uri: Path to config INI file

    :param sanity_check: Perform database sanity check on start

    :param console_app: Set true to setup console-mode logging. See :func:`setup_console_logging`

    :param extra_options: Passed through bootstrap() and is available as :attr:`websauna.system.Initializer.global_options`.

    :return: Faux Request object pointing to a site root, having registry and every configured.
    """

    # Paster thinks we are a string
    if sanity_check:
        sanity_check = "true"
    else:
        sanity_check = "false"

    options = {
        "sanity_check": sanity_check
    }

    if extra_options:
        options.update(extra_options)

    app = get_wsgi_app(config_uri, defaults=options)
    initializer = initializer_from_app(app)

    registry = initializer.config.registry
    dbsession = create_dbsession(registry)

    # Set up the request with websauna.site_url setting as the base URL
    request = make_routable_request(dbsession, registry)

    # This exposes the app object for the integration tests e.g test_static_asset
    # TODO: Find a cleaner way to do this
    request.app = app

    return request


def init_websauna_script_env(config_uri: str) -> dict:
    """Initialize Websauna WSGI application for a IPython notebook.

    :param config_uri: Path to config INI file

    :return: Dictionary of shell variables
    """

    options = {"sanity_check": False}
    app = get_wsgi_app(config_uri, defaults=options)

    initializer = initializer_from_app(app)

    registry = initializer.config.registry
    dbsession = create_dbsession(registry)

    pyramid_env = scripting.prepare(registry=app.initializer.config.registry)
    pyramid_env["app"] = app
    pyramid_env["initializer"] = initializer

    # Websauna specific
    # Set up the request with websauna.site_url setting as the base URL
    request = make_routable_request(dbsession, registry)
    pyramid_env["request"] = request
    pyramid_env["dbsession"] = dbsession

    return pyramid_env



"""Jinja template filters used in admin UI."""

# Pyramid
from jinja2 import contextfilter
from pyramid.renderers import render
from pyramid.threadlocal import get_current_request

# Websauna
from websauna.system.admin.interfaces import IAdmin
from websauna.system.core.breadcrumbs import get_breadcrumbs


@contextfilter
def admin_breadcrumbs(jinja_ctx, context, **kw):
    """Render admin interface top breadcrumbs bar."""

    if not context:
        return ""

    request = jinja_ctx.get('request') or get_current_request()
    current_view_name = jinja_ctx.get("current_view_name")
    current_view_url = request.url
    crumbs = get_breadcrumbs(context, request, root_iface=IAdmin, current_view_name=current_view_name, current_view_url=current_view_url)

    assert crumbs, "Could not get breadcrumbs for {}".format(context)

    if len(crumbs) == 1:
        return ""

    return render("templates/admin/breadcrumbs.html", dict(context=context, breadcrumbs=crumbs), request=request)



# Standard Library
import logging

# SQLAlchemy
import sqlalchemy
from sqlalchemy import inspect
from sqlalchemy.ext.declarative.clsregistry import _ModuleMarker
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.orm import Session


logger = logging.getLogger(__name__)


def is_sane_database(Base, session: Session):
    """Check whether the current database matches the models declared in model base.

    Currently we check that all tables exist with all columns. What is not checked

    * Column types are not verified

    * Relationships are not verified at all (TODO)

    :param Base: Declarative Base for SQLAlchemy models to check

    :param session: SQLAlchemy session bound to an engine

    :return: True if all declared models have corresponding tables and columns.
    """

    engine = session.get_bind()
    iengine = inspect(engine)

    errors = False

    # Get tables currently in the database
    tables = iengine.get_table_names()

    # Go through all SQLAlchemy models
    for name, klass in Base._decl_class_registry.items():

        if isinstance(klass, _ModuleMarker):
            # Not a model
            continue

        # 1. Check if this model declares any tables they exist

        # First try raw table definition
        __table__ = getattr(klass, "__table__", None)
        if __table__ is not None:
            table = __table__.name
        else:
            table = getattr(klass, "__tablename__", None)

        if not table:
            raise RuntimeError("Table definition missing for {}".format(table))

        if table not in tables:
            logger.error("Model %s declares table %s which does not exist in database %s", klass, table, engine)
            errors = True

        # 2. Check that all columns declared by the model exist
        mapper = inspect(klass)

        for column_prop in mapper.attrs:
            if isinstance(column_prop, RelationshipProperty):
                # TODO: Add sanity checks for relations
                continue

            # Iterate all columns the model declares
            for column in column_prop.columns:
                table_name = column.table.name

                # Get columns from the actual table the ORM referes to
                try:
                    columns = [c["name"] for c in iengine.get_columns(table_name)]
                except sqlalchemy.exc.NoSuchTableError:
                    logger.error("Model %s declares table %s which does not exist in database %s", klass, table_name, engine)
                    errors = True
                    break

                if column.key not in columns:
                    # It is safe to stringify engine where as password should be blanked out by stars
                    logger.error("Model %s declares column %s which does not exist in database %s", klass, column.key, engine)
                    errors = True

    return not errors



"""Define various interfaces telling how user subsystem objects interact and can be looked up from registry."""
# Pyramid
import zope
from pyramid.interfaces import IRequest
from pyramid.interfaces import IResponse
from zope.interface import Interface

import authomatic


class IUser(Interface):
    """User.

    Usually SQLAlchemy model instance of :py:class:`websauna.system.user.usermixin.UserMixin`.

    Hard requirements for User interface listed here - this is what Websauna default frontend expects from an user instance.

    :py:class:`websauna.system.user.interfaces.ILoginService` must know some user implementation details.
    """

    #: How we present the user's name to the user itself. Usually. Picks one of 1) full name if set 2) username if set 3) email.
    friendly_name = zope.interface.Attribute("friendly_name")


class IGroup(Interface):
    """User group.

    Usually SQLAlchemy model instance of :py:class:`websauna.system.user.usermixin.GroupMixin` but can be any object.
    """

    #: Then name of the group
    name = zope.interface.Attribute("name")


class IUserModel(Interface):
    """Register utility registration which marks active User SQLAlchemy model class."""


class IGroupModel(Interface):
    """Register utility registration which marks active Group SQLAlchemy model class."""


class IActivationModel(Interface):
    """Register utility registration which marks active Activation SQLAlchemy model class."""


class IAuthomatic(Interface):
    """Mark Authomatic instance in the registry."""


class ISocialLoginMapper(Interface):
    """Named marker interface to look up social login mappers."""

    def capture_social_media_user(self, request: IRequest, result: authomatic.core.LoginResult) -> IUserModel:
        """Extract social media information from the Authomatic login result in order to associate the user account."""

    def import_social_media_user(self, user: authomatic.core.User) -> dict:
        """Map incoming social network data to internal data structure.

        Sometimes social networks change how the data is presented over API and you might need to do some wiggling to get it a proper shape you wish to have.

        The resulting dict must be JSON serializable as it is persisted as is.

        :param user: Authomatic user.
        :returns: Dict representation of the user.
        """

    def update_first_login_social_data(self, user: object, data: dict):
        """Set the initial data on the user model.

        When the user logs in from a social network for the first time (no prior logins with this email before) we fill in blanks in the user model with incoming data.

        Default action is not to set any items.

        :param user: User object.
        :param data: Normalized data.
        """


class ISiteCreator(Interface):
    """Utility that is responsible to create the initial site."""


class AuthenticationFailure(Exception):
    """The user is not allowed to log in."""


class ILoginService(Interface):
    """A service that is responsible for handling normal website facing log in actions.

    This service is responsible to

        * Set up logged in session

        * Do post login actions like redirects

    Use :py:func:`websauna.system.user.utils.get_login_service` to get access to configured login service.
    """

    def authentication_user(user: IUser, login_source: str, location: str=None) -> IResponse:
        """Make the current session logged in session for this particular user.

        A password check is not performed. However it is checked if user is active and such.

        :param location: Override the redirect page. If none use ``websauna.login_redirect``. TODO - to be changed.
        :param login_source: Application specific string telling where the login come from. E.g. "social_media", "signup", "login_form".
        :raise AuthenticationFailure: If the user is disabled
        """

    def authenticate_credentials(username: str, login_source: str, password: str, location: str=None) -> IResponse:
        """Logs in the user.

        This is called after the user credentials have been validated.

        Sets the auth cookies and redirects to a post login page, which defaults to a view named 'index'.

        Fills in user last login time and IP data..

        :param request: Current request

        :param user: Default login service is designed to work with UserMixin compatible user classes

        :param location: Override the redirect page. If none use ``websauna.login_redirect``. TODO - to be changed.

        :param login_source: Application specific string telling where the login come from. E.g. "social_media", "signup", "login_form".

        :raise AuthenticationFailure: If the password does not match or user is disabled
        """

    def logout(location: str=None) -> IResponse:
        """Log out user from the site.

        * Terminate session

        * Show logged out message

        * Redirect the user to post login page
        """


class IOAuthLoginService(Interface):
    """A login service for federated authentication.

    See :py:class:`websauna.system.interfaces.ILoginService`.

    Use :py:func:`websauna.system.user.utils.get_oauth_login_service` to get access to configured login service.
    """

    def handle_request(provider_name: str) -> IResponse:
        """Handle all requests coming to login/facebook, login/twitter etc. endpoints.

        * Login form does an empty HTTP POST request to initiate OAuth process

        * Federated authentication service does HTTP GET redirect when they accept OAuth authentication request
        """


class IUserRegistry(Interface):
    """Manage creation and querying of users.

    Allow abstraction over the user backend - do not assume users are stored in the primary database.

    TODO: Interface not described yet, see :py:class:`websauna.system.user.userregistry.DefaultEmailBasedUserRegistry`.
    """


class CannotResetPasswordException(Exception):
    """Password reset is disabled for this user e.g. due to disabled account."""


class ICredentialActivityService(Interface):
    """User password and activation related activities.

    TODO: Interface not described yet, see :py:class:`websauna.system.user.credentialactivityservice.DefaultCredentialActivityService`.
    """


class IRegistrationService(Interface):
    """Sign up form service.

    TODO: Interface not described yet, see :py:class:`websauna.system.user.registrationservice.DefaultRegistrationService`.

    """


class ILoginSchema(Interface):
    """Colander schema used for sign in form.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class ILoginForm(Interface):
    """Deform form used for sign in form.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IRegisterSchema(Interface):
    """Colander schema used for sign upform.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IRegisterForm(Interface):
    """Deform form used for sign upform.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IForgotPasswordForm(Interface):
    """Deform form used for Forgot password form.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IForgotPasswordSchema(Interface):
    """Colander schema used for Forgot password form.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IResetPasswordForm(Interface):
    """Deform form used for Reset password form.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IResetPasswordSchema(Interface):
    """Colander schema used for Reset password form.

    See :py:meth:`websauna.system.Initializer.configure_user_forms`.
    """


class IPasswordHasher(Interface):
    """A utility for hashing passwords.

    Used by :py:meth:`websauna.system.models.usermixin.UserMixin._set_password`.
    """

    def hash_password(plain_text: str) -> str:
        """Generate a hash presentation for plain text password.

        This is to be stored in database.

        :return: A hasher internal string format. Usually contains number of cycles, hashed password and salt string.
        """

    def verify_password(hashed_password: str, plain_text: str) -> bool:
        """Verify a password.

        Compare if inputed password matches one stored in the dabase.

        :return: True if the password matches, False otherwise.
        """



"""Database default base models and session setup."""


# Pyramid
import transaction
import zope.sqlalchemy
from pyramid.registry import Registry

# SQLAlchemy
from sqlalchemy import engine_from_config
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import MetaData

# Websauna
from websauna.system.http import Request
from websauna.system.model.interfaces import ISQLAlchemySessionFactory

from .json import init_for_json
from .json import json_serializer


# Recommended naming convention used by Alembic, as various different database
# providers will autogenerate vastly different names making migrations more
# difficult. See: http://alembic.readthedocs.org/en/latest/naming.html
NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
}


metadata = MetaData(naming_convention=NAMING_CONVENTION)

#: This is a default SQLAlchemy model base class. Models inhering from this class are automatically registered with the default SQLAlchemy session. You can have your alternative Base class, but in this case you need to bind the Base class session yourself.
Base = declarative_base(metadata=metadata)


@event.listens_for(Base, "class_instrument", propagate=True)
def _on_model_registered(cls):
    """Intercept SQLAlchemy model creation (instrumentation).

    Insert event listeners that allow JSON values play nicely.
    """
    init_for_json(cls)


def includeme(config):
    """Hook up database initialization and SQLAlchemy global setup."""

    config.add_request_method(
        request_session_factory,
        'dbsession',
        reify=True
    )

    config.add_request_method(
        get_request_tm,
        'transaction_manager',
        reify=True
    )


def request_session_factory(request: Request) -> Session:
    """Look SQLAlchemy session creator."""
    session = request.registry.queryAdapter(request, ISQLAlchemySessionFactory)
    assert session, "No configured ISQLAlchemySessionFactory"
    return session


def get_request_tm(request: Request):
    """Get the underlying transaction manager."""
    return request.dbsession.transaction_manager


def create_transaction_manager_aware_dbsession(request: Request) -> Session:
    """Defaut database factory for Websauna.

    Looks up database settings from the INI and creates an SQLALchemy session based on the configuration. The session is terminated on the HTTP request finalizer.
    """
    dbsession = create_dbsession(request.registry, request.tm)

    def terminate_session(request):
        # Close db session at the end of the request and return the db connection back to the pool
        dbsession.close()

    request.add_finished_callback(terminate_session)

    return dbsession


def _get_psql_engin(settings, prefix):
    """Create PostgreSQL engine.

    The database engine defaults to SERIALIZABLE isolation level.
    :param settings:
    :param prefix:
    :return:
    """
    # http://stackoverflow.com/questions/14783505/encoding-error-with-sqlalchemy-and-postgresql
    engine = engine_from_config(settings, 'sqlalchemy.', connect_args={"options": "-c timezone=utc"}, client_encoding='utf8', isolation_level='SERIALIZABLE', json_serializer=json_serializer)
    return engine


def _get_sqlite_engine(settings, prefix):
    engine = engine_from_config(settings, 'sqlalchemy.', isolation_level='SERIALIZABLE')
    return engine


def get_engine(settings: dict, prefix='sqlalchemy.') -> Engine:
    """Reads config and create a database engine out of it."""
    url = settings.get("sqlalchemy.url")
    if not url:
        raise RuntimeError("sqlalchemy.url missing in the settings")

    if "sqlite" in url:
        return _get_sqlite_engine(settings, prefix)
    elif "postgres" in url:
        return _get_psql_engin(settings, prefix)
    else:
        raise RuntimeError("Unknown SQLAlchemy connection URL: {url}".format(url=url))


def create_session_maker(engine):
    """Create database session maker.

    Must be called only once per process.
    """
    dbmaker = sessionmaker()
    dbmaker.configure(bind=engine)
    return dbmaker


def create_dbsession(registry: Registry, manager=None) -> Session:
    """Creates a new database using the configured session poooling.

    This is called outside request life cycle when initializing and checking the state of the databases.

    :param manager: Transaction manager to bound the session. The default is thread local ``transaction.manager``.
    """

    assert isinstance(registry, Registry), "The first arg must be registry (Method signature changed)"

    # Make sure dbmaker is created only once per process as it must be
    # per-process for the connection pooling to work.
    # Cache the resulting object on Pyramid registry.
    db_session_maker = getattr(registry, "db_session_maker", None)

    if not db_session_maker:
        engine = get_engine(registry.settings)
        db_session_maker = registry.db_session_maker = create_session_maker(engine)

    if not manager:
        manager = transaction.manager

    dbsession = create_session(manager, db_session_maker)

    # Register zope.sqlalchemy extensino
    # register(dbsession, manager)

    return dbsession


def create_session(transaction_manager, db_session_maker: sessionmaker) -> Session:
    """Create a new database session with Zope transaction manager attached.

    The attached transaction manager takes care of committing the transaction at the end of the request.
    """
    dbsession = db_session_maker()
    transaction_manager.retry_attempt_count = 3  # TODO: Hardcoded for now
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    dbsession.transaction_manager = transaction_manager
    return dbsession



"""Test form autogeneration and CRUD."""
# Standard Library
import time

# Pyramid
import transaction

import pytest
from splinter.driver import DriverAPI

# Websauna
import websauna
from websauna.system.tests.utils import create_logged_in_user
from websauna.tests.webserver import customized_web_server
from websauna.utils.slug import slug_to_uuid
from websauna.utils.slug import uuid_to_slug


@pytest.fixture(scope="module")
def tutorial_app(request, paster_config):
    """Custom WSGI app with travesal points for sitemap enabled."""

    class Initializer(websauna.system.DemoInitializer):

        def run(self):
            super(Initializer, self).run()
            from websauna.system.tests import tutorial
            self.config.scan(tutorial)

    global_config, app_settings = paster_config
    init = Initializer(global_config, app_settings)
    init.run()
    app = init.make_wsgi_app()
    app.init = init
    return app


@pytest.fixture(scope="module")
def web_server(request, tutorial_app):
    """Run a web server with tutorial installed."""
    web_server = customized_web_server(request, tutorial_app)
    return web_server()


@pytest.fixture(scope="module")
def registry(request, tutorial_app):
    """Run a web server with tutorial installed."""
    return tutorial_app.init.config.registry


@pytest.mark.skip("Mark skipped for now before issues resolved on TravisCI - something to do with delays and browsers")
def test_add_question(browser: DriverAPI, registry, web_server, dbsession):
    """Adding questions should be succesful."""

    b = browser

    if b.driver.capabilities["browserName"] != "firefox":
        # Fails at click and JavaScript modals for Chrome
        pytest.skip("This test works only under Firefox WebDriver")

    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit(web_server)
    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-add-question").click()
    b.fill("question_text", "What is love")
    b.find_by_css("#deformField2-date").click()

    # Pick any date
    b.find_by_css(".picker__day--infocus")[0].click()

    time.sleep(0.8)  # Give some time for the browser, next click fails on CI

    b.find_by_css("#deformField2-time").click()
    b.find_by_css(".picker__list-item")[0].click()

    time.sleep(0.5)  # Give some time for the browser, next click fails on CI

    b.find_by_name("add").click()

    assert b.is_element_present_by_css("#msg-item-added")


def test_add_choice_no_question(browser: DriverAPI, registry, web_server, dbsession):
    """Add one choice, no questions available."""

    b = browser

    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit(web_server)
    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-add-choice").click()
    b.fill("choice_text", "Baby don't hurt me")
    b.find_by_name("add").click()

    assert b.is_element_present_by_css("#msg-item-added")


def test_add_choice_question(browser: DriverAPI, registry, web_server, dbsession):

    from .tutorial import Question
    from .tutorial import Choice

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()
        question_uuid = uuid_to_slug(q.uuid)

    b = browser

    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit(web_server)
    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-add-choice").click()
    b.fill("choice_text", "Baby don't hurt me")
    b.select("question", question_uuid)
    b.find_by_name("add").click()

    assert b.is_element_present_by_css("#msg-item-added")

    with transaction.manager:
        assert dbsession.query(Choice).first().question is not None


def test_add_choice_choose_no_question(browser: DriverAPI, registry, web_server, dbsession):

    from .tutorial import Question
    from .tutorial import Choice

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()

    b = browser
    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit(web_server)
    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-add-choice").click()
    b.fill("choice_text", "Baby don't hurt me")
    b.find_by_name("add").click()

    assert b.is_element_present_by_css("#msg-item-added")

    with transaction.manager:
        assert dbsession.query(Choice).first().question is None


def test_edit_choice_question(browser: DriverAPI, registry, web_server, dbsession):
    """Change choice's assigned question in edit."""

    from .tutorial import Question
    from .tutorial import Choice

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()

        q2 = Question(question_text="Who shot JFK")
        dbsession.add(q2)
        dbsession.flush()
        q2_slug = uuid_to_slug(q2.uuid)

        c = Choice(choice_text="Foobar", question=q)
        dbsession.add(c)
        dbsession.flush()
        c_slug = uuid_to_slug(c.uuid)

    b = browser
    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit("{}/admin/models/choice/{}/edit".format(web_server, c_slug))
    b.select("question", q2_slug)
    b.find_by_name("save").click()

    assert b.is_element_present_by_css("#msg-changes-saved")

    with transaction.manager:
        c = dbsession.query(Choice).get(1)
        assert c.question.uuid == slug_to_uuid(q2_slug)


def test_edit_choice_remove_question(browser: DriverAPI, registry, web_server, dbsession):
    """Editing choice allows us to reset question value back to null."""

    from .tutorial import Question
    from .tutorial import Choice

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()

        c = Choice(choice_text="Foobar", question=q)
        dbsession.add(c)
        dbsession.flush()
        c_slug = uuid_to_slug(c.uuid)

    b = browser
    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit("{}/admin/models/choice/{}/edit".format(web_server, c_slug))
    b.select("question", "")
    b.find_by_name("save").click()

    assert b.is_element_present_by_css("#msg-changes-saved")

    with transaction.manager:
        c = dbsession.query(Choice).get(1)
        assert c.question is None


def test_question_shows_choices(browser: DriverAPI, registry, web_server, dbsession):
    """If question has active choices they are shown on Show screen, albeit not editable."""

    from .tutorial import Question
    from .tutorial import Choice

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()
        q_slug = uuid_to_slug(q.uuid)

        c = Choice(choice_text="Baby don't hurt me", question=q)
        dbsession.add(c)
        dbsession.flush()

    b = browser
    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit("{}/admin/models/question/{}/show".format(web_server, q_slug))
    assert b.is_text_present("Baby don't hurt me")


def test_question_listing(browser: DriverAPI, registry, web_server, dbsession):
    """Question listing shows question text."""

    from .tutorial import Question

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()

    b = browser
    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit("{}/admin/models/question/listing".format(web_server))
    assert b.is_text_present("What is love")


def test_question_delete(browser: DriverAPI, registry, web_server, dbsession):
    """Delete question and make sure it deletes related choices.."""

    from .tutorial import Question
    from .tutorial import Choice

    with transaction.manager:
        q = Question(question_text="What is love")
        dbsession.add(q)
        dbsession.flush()

        c = Choice(choice_text="Baby don't hurt me", question=q)
        dbsession.add(c)
        dbsession.flush()
        q_slug = uuid_to_slug(q.uuid)

    b = browser
    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    b.visit("{}/admin/models/question/{}".format(web_server, q_slug))
    b.find_by_css("#btn-crud-delete").click()
    b.find_by_css("#btn-delete-yes").click()

    with transaction.manager:
        assert dbsession.query(Question).count() == 0
        assert dbsession.query(Choice).count() == 0



"""Admin interface events."""


class AdminConstruction:
    """Fired when admin interface is constructed for the request.

    Subscribers may contribute their own parts and traversing to admin UI.
    """

    def __init__(self, admin):
        self.admin = admin



# Standard Library
import time

# Pyramid
import transaction
from pyramid.interfaces import IAuthorizationPolicy
from pyramid.security import Everyone

import pytest
from flaky import flaky

# Websauna
from websauna.system.admin.utils import get_admin
from websauna.system.user.utils import get_site_creator
from websauna.system.tests.utils import EMAIL
from websauna.system.tests.utils import PASSWORD
from websauna.system.tests.utils import create_logged_in_user
from websauna.system.tests.utils import create_user


def test_admin_permissions(test_request):
    """Non-functional test to check admin permissions are sane."""

    admin = get_admin(test_request)
    policy = test_request.registry.queryUtility(IAuthorizationPolicy)

    # Admin group access ok
    assert policy.permits(admin, "group:admin", "view")
    assert policy.permits(admin, "group:admin", "edit")
    assert policy.permits(admin, "group:admin", "delete")

    # Block world
    assert not policy.permits(admin, Everyone, "view")
    assert not policy.permits(admin, Everyone, "edit")
    assert not policy.permits(admin, Everyone, "delete")


def test_enter_admin(web_server, browser, dbsession, init):
    """The first user can open the admin page."""

    with transaction.manager:
        u = create_user(dbsession, init.config.registry)
        site_creator = get_site_creator(init.config.registry)
        site_creator.init_empty_site(dbsession, u)
        assert u.is_admin()

    b = browser
    b.visit(web_server + "/login")

    b.fill("username", EMAIL)
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    assert b.is_element_visible_by_css("#nav-admin")
    b.find_by_css("#nav-admin").click()

    assert b.is_element_present_by_css("#admin-main")


def test_non_admin_user_denied(web_server, browser, dbsession, init):
    """The second user should not see admin link nor get to the admin page."""

    with transaction.manager:
        u = create_user(dbsession, init.config.registry, admin=True)
        assert u.is_admin()

        u = create_user(dbsession, init.config.registry, email="example2@example.com")
        assert not u.is_admin()

    b = browser
    b.visit(web_server + "/login")

    b.fill("username", "example2@example.com")
    b.fill("password", PASSWORD)
    b.find_by_name("login_email").click()

    assert not b.is_element_visible_by_css("#nav-admin")

    b.visit(web_server + "/admin/")
    assert b.is_element_visible_by_css("#forbidden")


@flaky
@pytest.mark.notebook
def test_context_sensitive_shell(web_server, browser, dbsession, init):
    """See we can open a context sensitive shell in admin."""

    if dbsession.bind.dialect.name == "sqlite":
        pytest.skip("This fails with sqlite on Travis - a fact that doesn't make sense, but it does")

    b = browser
    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#latest-user-shortcut").click()
    b.find_by_css("#btn-crud-shell").click()

    # Ramping up shell takes some extended time
    time.sleep(5)

    # We succesfully exposed obj
    assert b.is_text_present("example@example.com")

    # File menu
    b.find_by_css(".dropdown a")[0].click()

    # Shutdown and Back to the home
    assert b.is_element_visible_by_css("#shutdown")
    b.find_by_css("#shutdown").click()

    # There should be alert "Do you really wish to leave notebook?"
    time.sleep(0.5)
    alert = b.driver.switch_to_alert()
    alert.accept()

    # Back to home screen
    assert b.is_element_visible_by_css("#nav-logout")



"""Cross-site request forgery test views."""
# Pyramid
from pyramid.response import Response
from pyramid.view import view_config


@view_config(route_name="home")
def home(request):
    """We need one GET request to initialize session and set CSRF token."""
    return Response("OK")


@view_config(route_name="csrf_sample")
def csrf_sample(request):
    assert request.method == "POST"
    return Response("OK")


@view_config(route_name="csrf_exempt_sample", require_csrf=False)
def csrf_exempt_sample(request):
    assert request.method == "POST"
    return Response("OK")


@view_config(route_name="csrf_exempt_sample_context", require_csrf=False)
def csrf_exempt_sample_context(context, request):
    assert request.method == "POST"
    return Response("OK")



"""Duplicity based backuping."""
# Standard Library
import logging
import os
import stat
import subprocess

# Pyramid
from pyramid.path import AssetResolver
from pyramid.threadlocal import get_current_registry

# Websauna
from websauna.system.devop.exportenv import create_settings_env


__here__ = os.path.dirname(__file__)

logger = logging.getLogger(__name__)


def backup_site():
    """Run the site backup script.

    Runs the configured backup script ``websauna.backup_script``. The backup script can be any UNIX executable. The script gets the environment variables from *websauna* configuration, as exposed by :py:func:`websauna.utils.exportenv.create_settings_env`.

    In the case the backup script fails, an exception is raised and logged through normal application logging means.

    Run backup from the command line::

        echo "from websauna.system.devop import backup ; backup.backup_site()" | pyramid-web20-shell development.ini

    Note that the output is buffered, so there might not be instant feedback.
    """

    registry = get_current_registry()

    backup_script_spec = registry.settings.get("websauna.backup_script", "").strip()
    if not backup_script_spec:
        # Currently we do not have backup script defined, do not run
        return

    resolver = AssetResolver(None)
    backup_script = resolver.resolve(backup_script_spec).abspath()

    assert os.path.exists(backup_script), "Backup script does not exist: {}, spec {}".format(backup_script, backup_script_spec)

    assert stat.S_IXUSR & os.stat(backup_script)[stat.ST_MODE], "Backup script is not executable: {}".format(backup_script)

    backup_timeout = int(registry.settings.get("websauna.backup_timeout"))

    # Export all secrets and settings
    env = create_settings_env(registry)

    try:
        subprocess.check_output([backup_script, ], timeout=backup_timeout, stderr=subprocess.STDOUT, env=env)
    except subprocess.CalledProcessError as e:
        # Capture error to Sentry
        logger.error(e.output)
        raise



"""Typing support for Websauna."""
# flake8: noqa
# Standard Library
from posix import stat_result as stat_result
from typing import *  # pragma: no cover


class DirEntry(Generic[AnyStr]):  # noqa: F405
    # This is what the scandir interator yields
    # The constructor is hidden

    name = ''
    path = ''

    def inode(self) -> int: ...  # noqa: E704

    def is_dir(self, follow_symlinks: bool = ...) -> bool: ...  # noqa: E704

    def is_file(self, follow_symlinks: bool = ...) -> bool: ...  # noqa: E704

    def is_symlink(self) -> bool: ...  # noqa: E704

    def stat(self) -> stat_result: ...  # noqa: E704



"""Admin interface main views. """
# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from pyramid_layout.panel import panel_config

# Websauna
from websauna.system.admin.modeladmin import ModelAdmin
from websauna.system.admin.modeladmin import ModelAdminRoot
from websauna.system.admin.utils import get_admin
from websauna.system.core.panel import render_panel
from websauna.system.core.viewconfig import view_overrides
from websauna.system.crud import views as crud_views
from websauna.system.crud import listing
from websauna.system.crud.formgenerator import SQLAlchemyFormGenerator
from websauna.system.crud.sqlalchemy import sqlalchemy_deleter
from websauna.system.crud.views import TraverseLinkButton


@view_config(route_name='admin_home', renderer='admin/admin.html', permission='view')
def admin(request):
    """Admin front page page."""

    admin = get_admin(request)

    # For now, admin panels always appear in ascending order

    model_admin_root = admin["models"]

    # TODO: Have renderer adapters for panels, so that they can override views
    admin_panels = sorted(model_admin_root.items(), key=lambda pair: pair[1].title)
    rendered_panels = [render_panel(ma, request, name="admin_panel") for id, ma in admin_panels]

    return dict(panels=rendered_panels)


@panel_config(name='admin_panel', context=ModelAdmin, renderer='admin/model_panel.html')
def default_model_admin_panel(context, request, **kwargs):
    """Generic panel for any model admin.

    Display count of items in the database.
    """
    model_admin = context
    count = model_admin.get_query().count()
    admin = model_admin.__parent__
    title = model_admin.title
    return dict(locals(), **kwargs)


class Listing(crud_views.Listing):
    """Base listing view for all admin models.

    """
    base_template = "admin/base.html"

    table = listing.Table(
        columns=[
            listing.Column(id="id", name="Id",),
            listing.StringPresentationColumn(id="value", name="Value"),
            listing.ControlsColumn()
        ]
    )

    @property
    def title(self):
        return "All {}".format(self.context.title)

    @view_config(context=ModelAdmin, name="listing", renderer="crud/listing.html", route_name="admin", permission='view')
    def listing(self):
        # We override this method just to define admin route_name traversing
        return super(Listing, self).listing()


class Show(crud_views.Show):
    """Default show view for model admin."""
    base_template = "admin/base.html"

    form_generator = SQLAlchemyFormGenerator()

    resource_buttons = [
        TraverseLinkButton(id="edit", name="Edit", view_name="edit", permission="edit"),
        TraverseLinkButton(id="delete", name="Delete", view_name="delete", permission="delete"),
        TraverseLinkButton(id="shell", name="Shell", view_name="shell", permission="shell", tooltip="Open IPython Notebook shell and have this item prepopulated in obj variable.", feature="notebook"),
    ]

    @view_config(context=ModelAdmin.Resource, name="show", renderer="crud/show.html", route_name="admin", permission='view')
    def show(self):
        # We override this method just to define admin route_name traversing
        return super(Show, self).show()


class Edit(crud_views.Edit):
    """Default edit vie for model admin."""
    base_template = "admin/base.html"

    form_generator = SQLAlchemyFormGenerator()

    @view_config(context=ModelAdmin.Resource, name="edit", renderer="crud/edit.html", route_name="admin", permission='edit')
    def edit(self):
        # We override this method just to define admin route_name traversing
        return super(Edit, self).edit()


class Add(crud_views.Add):
    """Default add view for model admin."""
    base_template = "admin/base.html"

    form_generator = SQLAlchemyFormGenerator()

    @view_config(context=ModelAdmin, name="add", renderer="crud/add.html", route_name="admin", permission='add')
    def add(self):
        # We override this method just to define admin route_name traversing
        return super(Add, self).add()


@view_overrides(context=ModelAdmin.Resource, route_name="admin")
class Delete(crud_views.Delete):
    """Delete view for SQLAlchemy model admins."""

    base_template = "admin/base.html"

    deleter = sqlalchemy_deleter


@view_config(context=ModelAdmin, name="", route_name="admin", permission='view')
def model_admin_default_view(context, request):
    """Redirect to listing if model admin URL is being accessed without a view name."""
    return HTTPFound(request.resource_url(context, "listing"))


@view_config(context=ModelAdmin.Resource, name="", route_name="admin", permission='view')
def model_resource_default_view(context, request):
    """Redirect to show if model instance URL is being accessed without a view name."""
    return HTTPFound(request.resource_url(context, "show"))


@view_config(route_name='admin', context=ModelAdminRoot, permission='view')
def view__model_admin_root(context, request):
    """Model admin root does not have a view per se so we redirect to admin root."""
    return HTTPFound(request.resource_url(context.__parent__))



"""Extra widgets.

Mostly for high level integration.
"""
# Standard Library
import json

# Pyramid
import deform
from colander import null

# Websauna
from websauna.utils.slug import uuid_to_slug


class FriendlyUUIDWidget(deform.widget.TextInputWidget):
    """Display both UUID and base64 encoded string of a stored UUID value.

    For :py:class:`websauna.form.field.UUID` Colander type.
    """

    readonly_template = 'readonly/uuid'

    def get_template_values(self, field, cstruct, kw):
        values = {'cstruct': str(cstruct), 'field': field, 'slug': uuid_to_slug(cstruct) if cstruct else ''}
        values.update(kw)
        values.pop('template', None)
        return values


JSON_PREFORMATTED_STYLE = "font-family: monospace"


class JSONWidget(deform.widget.TextAreaWidget):
    """Nice JSON editor.

    Prettyprints outgoing JSON for better readability.

    Example usage:

    .. code-block:: python

        import colander

        from websauna.system.form.fields import JSONValue
        from websauna.system.form.widgets import JSONWidget

        class MySchema(colander.Schema):

            other_data = colander.SchemaNode(
                JSONValue(),
                widget=JSONWidget(),
                description="JSON bag of attributes of the object",
                missing=dict)

    """

    readonly_template = 'readonly/json'

    def __init__(self, cols=80, rows=10, style=JSON_PREFORMATTED_STYLE, prettyprint=True, **kwargs):
        super(JSONWidget, self).__init__(cols=cols, rows=rows, style=style, **kwargs)
        self.prettyprint = prettyprint

    def process_prettyprint(self, cstruct: str):
        if self.prettyprint:
            try:
                cstruct = json.loads(cstruct)
                return json.dumps(cstruct, sort_keys=True, indent=4)
            except Exception as e:
                pass

        return cstruct

    def serialize(self, field, cstruct, **kw):
        if cstruct in (null, None):
            cstruct = ''
        readonly = kw.get('readonly', self.readonly)
        template = readonly and self.readonly_template or self.template
        values = self.get_template_values(field, cstruct, kw)

        values["cstruct"] = self.process_prettyprint(values["cstruct"])
        return field.renderer(template, **values)



# Pyramid
from zope.interface import Interface


class IAdmin(Interface):
    """Marker interface for configured Admin class.     """


class IModelAdmin(Interface):
    """Marker interface for scanned model admins."""



"""JSONB field behavior checks."""
# Pyramid
import transaction

# Websauna
from websauna.system.model.json import NestedMutationDict
from websauna.system.user.models import User
from websauna.system.tests.utils import create_user


def test_pending_jsonb_dict_new_key(dbsession, registry):
    """Check that new keys added to JSONB that is not committed yet are persistent."""

    with transaction.manager:
        u = create_user(dbsession, registry)
        assert isinstance(u.user_data, NestedMutationDict)
        u.user_data["phone_number"] = "xxx"

    with transaction.manager:
        u = dbsession.query(User).first()
        assert u.user_data.get("phone_number") == "xxx"


def test_committed_jsonb_dict_new_key(dbsession, registry):
    """Check that new keys added to JSONB dict that is loaded from db are persistent."""

    with transaction.manager:
        u = create_user(dbsession, registry)
        print(u.user_data)

    with transaction.manager:
        u = dbsession.query(User).first()
        assert isinstance(u.user_data, NestedMutationDict)
        u.user_data["phone_number"] = "xxx"

    with transaction.manager:
        u = dbsession.query(User).first()
        assert u.user_data.get("phone_number") == "xxx"



# Pyramid
from pyramid.renderers import render
from pyramid.response import Response

import premailer

# Websauna
from websauna.system.core.route import simple_route


@simple_route("/sample-html-email", route_name="sample_html_email")
def sample_html_email(request):
    """Render a sample of HTML email.

    This view is conditionally enabled for development sites. See :ref:`websauna.sample_html_email`.
    """

    html_body = render("email/sample.html", {}, request=request)
    html_body = premailer.transform(html_body)

    return Response(html_body)



"""Map SQLAlchemy columns and other class attributes to Colander schema.

Also encapsulate :term:`colanderalchemy` so that we don't need to directly expose it in the case we want to get rid of it later.
"""
# Standard Library
import logging
import typing as t
from abc import ABC
from abc import abstractmethod

# Pyramid
import colander
import deform

# SQLAlchemy
from sqlalchemy import Column
from sqlalchemy import LargeBinary
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import UUID as PostgreSQLUUID
from sqlalchemy.dialects.postgresql import INET
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative.clsregistry import _class_resolver
from sqlalchemy.orm import Mapper
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql.type_api import TypeEngine

# Websauna
from websauna.system.crud import Resource
from websauna.system.form.colander import PropertyAwareSQLAlchemySchemaNode
from websauna.system.form.colander import TypeOverridesHandling
from websauna.system.form.fields import JSONValue
from websauna.system.form.sqlalchemy import UUIDForeignKeyValue
from websauna.system.form.sqlalchemy import UUIDModelSet
from websauna.system.form.sqlalchemy import get_uuid_vocabulary_for_model
from websauna.system.form.widgets import FriendlyUUIDWidget
from websauna.system.form.widgets import JSONWidget
from websauna.system.http import Request
from websauna.system.model import columns

from . import fields
from .editmode import EditMode


# TODO: Clean this up when getting rid of colanderalchemy
try:
    # Temporary support to suppress problematic form generation with geoalchemy
    from geoalchemy2 import Geometry
except ImportError:
    Geometry = ()


logger = logging.getLogger(__name__)


class ColumnToFieldMapper(ABC):
    """A helper class to map a SQLAlchemy model to Colander/Deform form."""

    @abstractmethod
    def map(self, mode: EditMode, request: Request, context: Resource, model: type, includes: t.List) -> colander.SchemaNode:
        """Map a model to a Colander form schema.

        :param mode: IS this add, edit or show form. For example, some relationship fields do not make sense on add form.

        :param context: Resource with ``get_object()`` to get the actual SQLAlchemy object

        :param model: SQLAlchemy model class

        :param includes: List of [column name | SchemaNode] we need to map

        :return: colander.SchemaNode which presents the automatically generated form
        """


class DefaultSQLAlchemyFieldMapper(ColumnToFieldMapper):
    """The default SQLAlchemy to form field and widget mapping implementation.

    We support

    * The default colanderalchemy mappings

    * UUID

    * JSONProperty declarations

    * IP addresses (INET)

    * ForeignKey references where it holds a reference to one another SQLAlchemy object

    See :py:class:`colanderalchemy.schema.SQLAlchemySchemaNode` for more information.
    """

    def map_standard_relationship(self, mode, request, node, model, name, rel) -> colander.SchemaNode:
        """Build a widget for choosing a relationship with target.

        The relationship must be foreign_key and the remote must offer ``uuid`` attribute which we use as a vocabulary key..
        """

        if isinstance(rel.argument, Mapper):
            # TODO: How to handle this kind of relationships
            return TypeOverridesHandling.drop

        if isinstance(rel.argument, _class_resolver):
            # Case from tutorialapp:
            # <RelationshipProperty at 0x1095a4438; question>
            # <class 'sqlalchemy.ext.declarative.clsregistry._class_resolver'>
            remote_model = rel.argument()
        else:
            remote_model = rel.argument

        # Get first column of the set
        for column in rel.local_columns:
            break

        # For now, we automatically deal with this only if the model provides uuid
        if hasattr(remote_model, "uuid"):
            dbsession = request.dbsession
            # TODO: We probably need a mechanism for system wide empty default label

            required = not column.nullable

            if mode in (EditMode.add, EditMode.edit):
                default_choice = "--- Choose one ---"
            else:
                default_choice = "(not set)"

            if required:
                missing = colander.required
            else:
                missing = None

            vocabulary = get_uuid_vocabulary_for_model(dbsession, remote_model, default_choice=default_choice)

            if rel.uselist:
                # Show out all relationships
                if mode == EditMode.show:
                    return colander.SchemaNode(UUIDModelSet(remote_model), name=name, missing=missing, widget=deform.widget.CheckboxChoiceWidget(values=vocabulary))
            else:
                # Select from a single relationship
                return colander.SchemaNode(UUIDForeignKeyValue(remote_model), name=name, missing=missing, widget=deform.widget.SelectWidget(values=vocabulary))

        return TypeOverridesHandling.drop

    def map_relationship(self, mode: EditMode, request: Request, node: colander.SchemaNode, model: type, name: str, rel: RelationshipProperty, mapper: Mapper):

        # Ok this is something we can handle, a single reference to another
        return self.map_standard_relationship(mode, request, node, model, name, rel)

    def map_column(self, mode: EditMode, request: Request, node: colander.SchemaNode, model: type, name: str, column: Column, column_type: TypeEngine) -> t.Tuple[colander.SchemaType, dict]:
        """Map non-relationship SQLAlchemy column to Colander SchemaNode.

        :return: Tuple(constructed colander.SchemaType, dict of addtional colander.SchemaNode construction arguments)
        """
        logger.debug("Mapping field %s, mode %s, node %s, column %s, column type %s", name, mode, node, column, column_type)

        # Check for autogenerated columns (updated_at)
        if column.onupdate:
            if mode in (EditMode.edit, EditMode.add):
                return TypeOverridesHandling.drop, {}

        # Don't fill default values when added, as they are automatically populated
        if column.default:
            if mode == EditMode.add:
                return TypeOverridesHandling.drop, {}

        # Never add primary keys
        # NOTE: TODO: We need to preserve ids because of nesting mechanism and groupedit widget wants it id
        if column.primary_key:
            # TODO: Looks like column.autoincrement is set True by default, so we cannot use it here
            if mode in (EditMode.edit, EditMode.add):
                return TypeOverridesHandling.drop, {}

        if column.foreign_keys:

            # Handled by relationship mapper
            return TypeOverridesHandling.drop, {}

        elif isinstance(column_type, (PostgreSQLUUID, columns.UUID)):

            # UUID's cannot be22 edited
            if mode in (EditMode.add, EditMode.edit):
                return TypeOverridesHandling.drop, {}

            # But let's show them
            return fields.UUID(), dict(missing=colander.drop, widget=FriendlyUUIDWidget(readonly=True))

        elif isinstance(column_type, Text):
            return colander.String(), dict(widget=deform.widget.TextAreaWidget())
        elif isinstance(column_type, JSONB):
            return JSONValue(), dict(widget=JSONWidget())
        elif isinstance(column_type, (JSONB, columns.JSONB)):
            # Can't edit JSON
            if mode in (EditMode.add, EditMode.edit):
                return TypeOverridesHandling.drop, {}
            return colander.String(), {}
        elif isinstance(column_type, LargeBinary):
            # Can't edit binary
            return TypeOverridesHandling.drop, {}
        elif isinstance(column_type, Geometry):
            # Can't edit geometry
            return TypeOverridesHandling.drop, {}
        elif isinstance(column_type, (INET, columns.INET)):
            return colander.String(), {}
        else:
            # Default mapping / unknown, let the parent handle
            return TypeOverridesHandling.unknown, {}

    def map(self, mode: EditMode, request: Request, context: t.Optional[Resource], model: type, includes: t.List, nested=None) -> colander.SchemaNode:
        """
        :param mode: What kind of form is this - show, add, edit
        :param request: HTTP request
        :param context: Current traversing context or None
        :param model: The SQLAlchemy model class for which we generate schema
        :param includes: List of column, relationship and property names or ``colander.SchemaNode(name=name)`` instances to be included on the form.
        :param nested: Legacy. Going away.
        """

        def _map_column(node, name, column, column_type):
            return self.map_column(mode, request, node, model, name, column, column_type)

        def _map_relationship(node: colander.SchemaNode, name: str, prop: RelationshipProperty, mapper: Mapper):
            return self.map_relationship(mode, request, node, model, name, prop, mapper)

        # TODO: We need to get rid of nested
        # Don't try to pull in relatinoships on add form
        nested = True

        # TODO: Make explicitly passed
        dbsession = request.dbsession

        schema = PropertyAwareSQLAlchemySchemaNode(model, dbsession, includes=includes, type_overrides=_map_column, relationship_overrides=_map_relationship, automatic_relationships=True, nested=nested)
        return schema



# Standard Library
import enum
import json
import typing as t

# Pyramid
import colander
import deform

# Websauna
from websauna.system.model.json import NestedMutationDict
from websauna.system.model.json import NestedMutationList
from websauna.system.model.json import json_serializer


def defer_widget_values(widget: type, values_callback: t.Callable, **kwargs) -> deform.widget.Widget:
    """Allow select or checkbox widget values construction deferred during the execution time.

    :param widget: Any Deform widget class, see :py:class:`deform.widget.Widget`
    :param value_callback: This callable(node, kw) is called deferredly by Colander
    :param kwargs: Passed to the widget constructed
    """

    _widget = widget

    @colander.deferred
    def _inner(node, kw):
        widget = _widget(values=values_callback(node, kw))
        return widget

    return _inner


class UUID(colander.String):
    """UUID field for Colander.

    See also :py:class`websauna.system.form.widgets.FriendlyUUIDWidget`.
    """

    def serialize(self, node, appstruct):
        # Assume widgets can handle raw UUID object
        return appstruct


class EnumValue(colander.String):
    """Allow choice of python enum.Enum in colander schemas.

    Example:

    .. code-block:: python

        class AssetClass(enum.Enum):
        '''What's preferred display format for this asset.'''

            fiat = "fiat"
            cryptocurrency = "cryptocurrency"
            token = "token"
            tokenized_shares = "tokenized_shares"
            ether = "ether"

        class Schema(CSRFSchema):

            asset_class = colander.SchemaNode(
                EnumValue(AssetClass),
                widget=deform.widget.SelectWidget(values=enum_values(AssetClass)))

    """

    def __init__(self, enum_class: type):
        super().__init__()
        assert issubclass(enum_class, enum.Enum), "Expected Enum, got {}".format(enum_class)
        self.enum_class = enum_class

    def deserialize(self, node: colander.SchemaNode, cstruct: str):
        """Parse incoming form values to Python objects if needed.
        """
        if cstruct:
            return self.enum_class(cstruct)
        else:
            return None

    def serialize(self, node: colander.SchemaNode, _enum: enum.Enum) -> str:
        """Convert Enum object to str for widget processing."""
        if _enum:
            assert isinstance(_enum, self.enum_class), "Expected {}, got {}".format(self.enum_class, _enum)
            return _enum.value
        else:
            return _enum


class JSONValue(colander.String):
    """Serialize / deserialize JSON fields.

    Example:

    .. code-block:: python

        class AssetSchema(CSRFSchema):

            name = colander.SchemaNode(colander.String())

            other_data = colander.SchemaNode(
                JSONValue(),
                widget=JSONWidget(),
                description="JSON bag of attributes of the object")

    """

    def deserialize(self, node: colander.SchemaNode, cstruct: str):
        """Parse incoming form values to Python objects if needed.
        """
        if cstruct:
            try:
                return json.loads(cstruct)
            except json.JSONDecodeError as e:
                raise colander.Invalid(node, "Not valid JSON") from e
        else:
            return None

    def serialize(self, node: colander.SchemaNode, data: t.Union[list, dict]) -> str:
        """Convert Python objects to JSON string."""
        if data:
            assert isinstance(data, (list, dict, NestedMutationDict, NestedMutationList)), "Expected list or dict, got {}".format(data.__class__)
            return json_serializer(data)
        else:
            # Noneish
            return data



"""Automatic admin and CRUD for SQLAlchemy models."""
# Standard Library
import string
import typing as t

# Pyramid
import venusian
from pyramid.interfaces import IRequest

# Websauna
from websauna.system.admin.interfaces import IAdmin
from websauna.system.admin.interfaces import IModelAdmin
from websauna.system.core.traversal import Resource
from websauna.system.crud.sqlalchemy import CRUD as CRUD
from websauna.system.crud.sqlalchemy import Resource as AlchemyResource


# We enforce some best practices to readable URL names of model admins. This is an arbitrary choice of the author.
ALLOWED_TRAVERSE_ID_CHARACTERS = string.ascii_lowercase + string.digits + "-"


class ModelAdmin(CRUD):
    """Resource presenting one model in admin interface.

    Provide automatized list, show add, edit and delete actions for an SQLAlchemy model which declares admin interface.
    """

    #: Title used in breadcrumbs, other places
    title = None

    #: Model must be set by subclass
    model = None

    def __init__(self, request):
        super(ModelAdmin, self).__init__(request)

    def get_model(self):
        assert self.model, "Model must be set by a subclass as a class attribute"
        return self.model

    #: Our resource factory
    class Resource(AlchemyResource):

        def get_admin(self):
            return self.__parent__.get_admin()

    def get_admin(self) -> IAdmin:
        """Get Admin resource object."""
        return self.__parent__.__parent__

    def get_title(self) -> str:
        if self.title:
            return self.title
        return self.id.capitalize()


class ModelAdminRoot(Resource):
    """Admin resource under which all model admins lurk.

    To get access this resource:


    """

    def get_title(self):
        return "Models"

    def get_model_admins(self):
        """List all registered model admin classes.

        :yield: (model_id, IModelAdmin) tuples
        """

        for model_id, model_cls in self.request.registry.getAdapters([self.request], IModelAdmin):
            yield(model_id, model_cls)

    def __getitem__(self, item):
        """Traverse to model admins. """
        registry = self.request.registry
        model_admin_resource = registry.queryAdapter(self.request, IModelAdmin, name=item)
        if not model_admin_resource:
            raise RuntimeError("Did not find model admin with id: {}".format(item))

        Resource.make_lineage(self, model_admin_resource, item)
        return model_admin_resource

    def items(self) -> t.List[t.Tuple[str, ModelAdmin]]:
        for id, model_cls in self.get_model_admins():
            yield id, self[id]


def model_admin(traverse_id: str) -> type:
    """Class decorator to mark the class to become part of model admins.

    ``Configure.scan()`` must be run on this module for the implementation to be picked up.

    If there is already an existing model admin with same ``model``, then the existing model admin is overwritten.

    :param traverse_id: Under which URL id this model appears in the admin interface. Allowed to contain lowercase letters, dash and digits. This will be available as ``ModelAdmin.__name__`` instance attribute.

    :param model_cls: Which model class this admin resource is controlling
    """

    assert all(c in ALLOWED_TRAVERSE_ID_CHARACTERS for c in traverse_id), "traverse_id may only contain lowercase letters, digits and a dash: {}".format(traverse_id)

    def _inner(cls):
        "The class decorator example"

        def register(scanner, name, wrapped):
            config = scanner.config
            # We can look up midels by

            model = getattr(cls, "model", None)
            assert model, "Class {} must declare model attribute".format(cls)

            registry = config.registry

            # Purge existing model admin
            registry.unregisterAdapter(required=(IRequest,), provided=IModelAdmin)
            registry.registerAdapter(cls, required=(IRequest,), provided=IModelAdmin, name=traverse_id)
            registry.model_admin_ids_by_model[model] = traverse_id

        venusian.attach(cls, register, category='websauna')
        return cls

    return _inner


def configure_model_admin(config):
    """Sets up model -> model admin registry."""
    config.registry.model_admin_ids_by_model = {}



"""CRUD views for user and group management."""
# Pyramid
import colander
import deform
from pyramid.httpexceptions import HTTPFound
from pyramid.request import Request
from pyramid.request import Response
from pyramid.view import view_config
from pyramid_layout.panel import panel_config

# SQLAlchemy
from sqlalchemy.orm import Query

# Websauna
from websauna.system.admin import views as admin_views
from websauna.system.admin.utils import get_admin_url_for_sqlalchemy_object
from websauna.system.core import messages
from websauna.system.core.viewconfig import view_overrides
from websauna.system.crud import listing
from websauna.system.crud.formgenerator import SQLAlchemyFormGenerator
from websauna.system.crud.sqlalchemy import Resource
from websauna.system.crud.sqlalchemy import sqlalchemy_deleter
from websauna.system.crud.views import CSVListing
from websauna.system.crud.views import TraverseLinkButton
from websauna.system.form.fieldmapper import EditMode
from websauna.system.form.fields import JSONValue
from websauna.system.form.fields import defer_widget_values
from websauna.system.form.widgets import JSONWidget
from websauna.system.user import events
from websauna.system.user.admins import GroupAdmin
from websauna.system.user.admins import UserAdmin
from websauna.system.user.interfaces import IPasswordHasher
from websauna.system.user.interfaces import IUser
from websauna.system.user.models import User
from websauna.system.user.schemas import GroupSet
from websauna.system.user.schemas import group_vocabulary
from websauna.system.user.schemas import validate_unique_user_email
from websauna.system.user.utils import get_user_registry
from websauna.utils.time import now


def kill_user_sessions(request: Request, user: IUser, operation: str):
    """Notify session to drop this user.

    :param request: Pyramid request.
    :param user: User.
    :param operation: Operation triggering the killing of user sessions.
    """
    user.last_auth_sensitive_operation_at = now()
    e = events.UserAuthSensitiveOperation(request, user, operation)
    request.registry.notify(e, request)


@panel_config(name='admin_panel', context=UserAdmin, renderer='admin/user_panel.html')
def user_admin_panel(context: UserAdmin, request: Request, **kwargs) -> dict:
    """Admin panel for Users.

    :param context: Model admin.
    :param request: Pyramid request.
    :param kwargs: Additional context to be passed to the panel.
    :return: Context for template rendering.
    """
    dbsession = request.dbsession

    model_admin = context
    admin = model_admin.get_admin()
    model = model_admin.get_model()

    title = model_admin.title
    count = dbsession.query(model).count()
    latest_user = dbsession.query(model).order_by(model.id.desc()).first()
    latest_user_url = get_admin_url_for_sqlalchemy_object(admin, latest_user)

    return dict(locals(), **kwargs)


class UserListing(admin_views.Listing):
    """Listing view for Users."""
    title = "All users"

    table = listing.Table(
        columns=[
            listing.Column("id", "Id",),
            listing.Column("friendly_name", "Friendly name"),
            listing.Column("email", "Email"),
            listing.ControlsColumn()
        ]
    )

    # Include our CSV export button in the user interface
    resource_buttons = admin_views.Listing.resource_buttons + [
        TraverseLinkButton(id="csv-export", name="CSV Export", view_name="csv-export", permission="view")
    ]

    def order_query(self, query: Query) -> Query:
        """Add ordering to an existing SQLAlchemy query.

        :param query: Base query.
        :return: Query ordered by created_at.
        """
        return query.order_by(self.get_model().created_at.desc())

    @view_config(context=UserAdmin, route_name="admin", name="listing", renderer="crud/listing.html", permission='view')
    def listing(self) -> dict:
        """User listing view.

        :return: Context for template rendering.
        """
        return super(UserListing, self).listing()


def _get_user_navigate_target(view: CSVListing, column: listing.Column, obj: User) -> str:
    """Return the admin url to an User object.

    :param view: Current view.
    :param column: Listing Column
    :param obj: User object.
    :return: Admin url to the User object.
    """
    request = view.request
    admin = request.admin
    return get_admin_url_for_sqlalchemy_object(admin, obj)


@view_overrides(context=UserAdmin, route_name="admin")
class UserCSVListing(CSVListing):
    """CSV export of the site users."""

    title = "users-export"

    table = listing.Table(
        columns=[
            listing.Column("id"),
            listing.Column("uuid"),
            listing.Column("link", getter=_get_user_navigate_target),
            listing.Column("enabled"),
            listing.Column("created_at"),
            listing.Column("last_login_at"),
            listing.Column("username"),
            listing.Column("email"),
            listing.Column("friendly_name"),
        ]
    )

    def order_query(self, query: Query) -> Query:
        """Add ordering to an existing SQLAlchemy query.

        :param query: Base query.
        :return: Query ordered by created_at.
        """
        return query.order_by(self.get_model().created_at.desc())


class UserShow(admin_views.Show):
    """Show one user."""

    resource_buttons = admin_views.Show.resource_buttons + [TraverseLinkButton(id="set-password", name="Set password", view_name="set-password")]

    includes = [
        "id",
        "uuid",
        "enabled",
        "created_at",
        "updated_at",
        "username",
        colander.SchemaNode(colander.String(), name='full_name'),
        "email",
        "last_login_at",
        "last_login_ip",
        colander.SchemaNode(colander.String(), name="registration_source", missing=colander.drop),
        colander.SchemaNode(JSONValue(), name="user_data", widget=JSONWidget(), description="user_data JSON properties"),
        colander.SchemaNode(GroupSet(), name="groups", widget=defer_widget_values(deform.widget.CheckboxChoiceWidget, group_vocabulary, css_class="groups"))
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    def get_title(self) -> str:
        """Title for the User object.

        :return: Title for the User object.
        """
        user = self.get_object()
        return "{friendly_name} #{id}".format(
            friendly_name=user.friendly_name,
            id=self.get_object().id
        )

    @view_config(context=UserAdmin.Resource, route_name="admin", name="show", renderer="crud/show.html", permission='view')
    def show(self):
        """User show view.

        :return: Context for template rendering.
        """
        return super(UserShow, self).show()


class UserEdit(admin_views.Edit):
    """Edit one user in admin interface."""

    includes = [
        "enabled",
        colander.SchemaNode(colander.String(), name='username'),  # Make username required field
        colander.SchemaNode(colander.String(), name='full_name', missing=""),
        "email",
        colander.SchemaNode(GroupSet(), name="groups", widget=defer_widget_values(deform.widget.CheckboxChoiceWidget, group_vocabulary, css_class="groups"))
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    def save_changes(self, form: deform.Form, appstruct: dict, user: User):
        """Save the user edit and reflect if we need to drop user session.

        :param form: Form object.
        :param appstruct: Form data.
        :param user: User object.
        """
        enabled_changes = appstruct["enabled"] != user.enabled
        email_changes = appstruct["email"] != user.email
        username_changes = appstruct["username"] != user.username

        super(UserEdit, self).save_changes(form, appstruct, user)

        # Notify authentication system to drop all sessions for this user
        if enabled_changes:
            kill_user_sessions(self.request, user, "enabled_change")
        elif email_changes:
            kill_user_sessions(self.request, user, "email_change")
        elif username_changes:
            kill_user_sessions(self.request, user, "username_change")

    def get_title(self) -> str:
        """Title for the User object.

        :return: Title for the User object.
        """
        user = self.get_object()
        return "{friendly_name} #{id}".format(
            friendly_name=user.friendly_name,
            id=self.get_object().id
        )

    @view_config(context=UserAdmin.Resource, route_name="admin", name="edit", renderer="crud/edit.html", permission='edit')
    def edit(self):
        """User edit view.

        :return: Context for template rendering.
        """
        return super(UserEdit, self).edit()


@view_overrides(context=UserAdmin)
class UserAdd(admin_views.Add):
    """CRUD for creating new users."""
    #: TODO: Not sure how we should manage with explicit username - it's not used for login so no need to have a point to ask

    includes = [
        # "username", --- usernames are never exposed anymore
        colander.SchemaNode(colander.String(), name="email", validator=validate_unique_user_email),
        "full_name",
        colander.SchemaNode(colander.String(), name='password', widget=deform.widget.CheckedPasswordWidget(css_class="password-widget")),
        colander.SchemaNode(GroupSet(), name="groups", widget=defer_widget_values(deform.widget.CheckboxChoiceWidget, group_vocabulary, css_class="groups"))
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    def get_form(self):
        """Return the Add form for this view.

        :return: Form object.
        """
        # TODO: Still not sure how handle nested values on the automatically generated add form. But here we need it for groups to appear
        return self.create_form(EditMode.add, buttons=("add", "cancel",))

    def initialize_object(self, form: deform.Form, appstruct: dict, obj: User):
        """Initialize User object.

        :param form: Form object.
        :param appstruct: Form data.
        :param obj: User object.
        """
        password = appstruct.pop("password")
        form.schema.objectify(appstruct, obj)
        hasher = self.request.registry.getUtility(IPasswordHasher)
        obj.hashed_password = hasher.hash_password(password)

        # Users created through admin are useable right away, so activate the user
        obj.activated_at = now()


class UserSetPassword(admin_views.Edit):
    """Set the user password.

    Use the CRUD edit form with one field to set the user password.
    """

    includes = [
        colander.SchemaNode(colander.String(), name='password', widget=deform.widget.CheckedPasswordWidget(css_class="password-widget")),
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    def save_changes(self, form: deform.Form, appstruct: dict, obj: User):
        """Save the form data.

        :param form: Form object.
        :param appstruct: Form data.
        :param user: User object.
        """
        # Set hashed password
        user_registry = get_user_registry(self.request)
        user_registry.set_password(obj, appstruct["password"])

        # Drop session
        kill_user_sessions(self.request, obj, "password_change")

    def do_success(self) -> Response:
        """After password change, redirect user.

        :return: Redirect user.
        """
        messages.add(self.request, kind="success", msg="Password changed.", msg_id="msg-password-changed")
        # Redirect back to view page after edit page has succeeded
        return HTTPFound(self.request.resource_url(self.context, "show"))

    @view_config(context=UserAdmin.Resource, route_name="admin", name="set-password", renderer="crud/edit.html", permission='edit')
    def set_password(self):
        """User set password view.

        :return: Context for template rendering.
        """
        return super(admin_views.Edit, self).edit()


@view_overrides(context=UserAdmin.Resource)
class UserDelete(admin_views.Delete):
    """Delete user view.

    Drop user sessions on invocation.
    """
    def deleter(self, context: Resource, request: Request):
        """Execute user deletion.

        * Delete user from database.
        * Remove all user sessions.

        :param context: Traversable resource.
        :param request: Pyramid request.
        """
        sqlalchemy_deleter(self, context, request)
        kill_user_sessions(request, context.get_object(), "user_deleted")


@view_overrides(context=GroupAdmin)
class GroupListing(admin_views.Listing):
    """Listing view for Groups."""

    table = listing.Table(
        columns=[
            listing.Column("id", "Id",),
            listing.Column("name", "Name"),
            listing.Column("description", "Description"),
            listing.ControlsColumn()
        ]
    )

    def order_query(self, query: Query) -> Query:
        """Add ordering to an existing SQLAlchemy query.

        :param query: Base query.
        :return: Query ordered by id descending.
        """
        return query.order_by(self.get_model().id.desc())


class GroupShow(admin_views.Show):
    """Display one Group."""

    includes = [
        "id",
        "name",
        "description",
        "created_at",
        "updated_at"
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    @view_config(context=GroupAdmin.Resource, route_name="admin", name="show", renderer="crud/show.html", permission='view')
    def show(self):
        """Group show view..

        :return: Context for template rendering.
        """
        return super(GroupShow, self).show()


class GroupAdd(admin_views.Add):
    """Create new Group."""

    includes = [
        "name",
        "description"
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    @view_config(context=GroupAdmin, route_name="admin", name="add", renderer="crud/add.html", permission='add')
    def add(self):
        """Group add view..

        :return: Context for template rendering.
        """
        return super(GroupAdd, self).add()


class GroupEdit(admin_views.Edit):
    """Edit one group in admin interface."""

    includes = [
        "name",
        "description"
    ]

    form_generator = SQLAlchemyFormGenerator(includes=includes)

    @view_config(context=GroupAdmin.Resource, route_name="admin", name="edit", renderer="crud/edit.html", permission='edit')
    def edit(self):
        """Group edit view..

        :return: Context for template rendering.
        """
        return super(GroupEdit, self).edit()



"""Highly modified colanderalchemy core.

.. note ::

    This is a temporary solution and will be replaced with something cleaner after working with the upstream author.

"""

# Standard Library
import itertools
import logging

# Pyramid
import colander
from colander import Mapping
from colander import SchemaNode
from colander import Sequence
from colander import drop
from colander import required

# SQLAlchemy
import sqlalchemy
from sqlalchemy import Boolean
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import Enum
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import String
from sqlalchemy import Time
from sqlalchemy import inspect
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.schema import Column
from sqlalchemy.schema import ColumnDefault
from sqlalchemy.schema import FetchedValue

from colanderalchemy.schema import SQLAlchemySchemaNode
from colanderalchemy.schema import _creation_order

# Websauna
from websauna.system.form.sqlalchemy import ModelSchemaType
from websauna.system.form.sqlalchemy import ModelSetResultList
from websauna.utils.jsonb import is_index_property


log = logger = logging.getLogger(__name__)


class TypeOverridesHandling(Enum):

    #: Return value from type_overrides signaling that this column should not appear on the form
    drop = 1

    #: Type overrides don't want to deal with this column let it proceed as is
    unknown = 2


class PropertyAwareSQLAlchemySchemaNode(SQLAlchemySchemaNode):
    """ColanderAlchemy mapper with some extensions.

    Allow automatically map of JSON'ed properties besides SQL Alchemy columns.

    Allow ``type_overrides``.

    .. note ::

        The original function is the most well written horrifying piece of Python code I have seen. The author has good intentions and has build a fantastic extension mechanism which is mostly useless for us. Furthermore the original code functions are so long that they are very fragile to change via subclassing. Overriding a colander type for SQLAlchemy column type was impossible and we patch our own extension mechanism in here.

        TODO: Resolve the issue with the colanderalchemy author politically correct way. What we are doing now is just a temporary 0.1 solution.
    """

    def __init__(self, class_, dbsession, includes=None, excludes=None, overrides=None, unknown='ignore', nested=False, type_overrides=None, relationship_overrides=None, automatic_relationships=False, **kw):
        """
        :param includes:
        :param overrides:
        :param unknown:
        :param type_overrides: callable(name, column, column_type)
        :param automatic_relationships: Only follow relationships if they are on include list
        :param kw:
        :return:
        """
        assert type_overrides, "Always provide type_overrides, otherwise this crap doesn't work when these nodes get nested"

        self.inspector = inspect(class_)
        kwargs = kw.copy()

        # Obtain configuration specific from the mapped class
        kwargs.update(getattr(self.inspector.class_, self.ca_class_key, {}))
        unknown = kwargs.pop('unknown', unknown)

        # The default type of this SchemaNode is Mapping.
        super(SQLAlchemySchemaNode, self).__init__(Mapping(unknown), **kwargs)
        self.class_ = class_
        self.dbsession = dbsession
        self.includes = includes or {}
        self.excludes = excludes or {}
        self.overrides = overrides or {}
        self.unknown = unknown
        self.declarative_overrides = {}
        self.kwargs = kwargs or {}
        self.type_overrides = type_overrides
        self.relationship_overrides = relationship_overrides
        self.nested = nested
        self.automatic_relationships = automatic_relationships
        self.add_nodes(self.includes, self.excludes, self.overrides, nested)

    def add_nodes(self, includes, excludes, overrides, nested):

        if set(excludes) & set(includes):
            msg = 'excludes and includes are mutually exclusive.'
            raise ValueError(msg)

        properties = sorted(self.inspector.attrs, key=_creation_order)
        # sorted to maintain the order in which the attributes
        # are defined
        for name in includes or [item.key for item in properties]:
            prop = self.inspector.attrs.get(name, name)

            if name in excludes or (includes and name not in includes):
                log.debug('Attribute %s skipped imperatively', name)
                continue

            name_overrides_copy = overrides.get(name, {}).copy()

            if (isinstance(prop, ColumnProperty) and isinstance(prop.columns[0], Column)):
                node = self.get_schema_from_column(
                    prop,
                    name_overrides_copy
                )
            elif isinstance(prop, RelationshipProperty):

                if not nested:
                    log.debug('Attribute %s skipped because not recursing to nested', name)
                    continue

                node = self.get_schema_from_relationship(
                    prop,
                    name_overrides_copy
                )
            elif isinstance(prop, colander.SchemaNode):
                node = prop
            else:
                log.debug(
                    'Attribute %s skipped due to not being '
                    'a ColumnProperty or RelationshipProperty',
                    name
                )
                continue

            if node is not None:
                self.add(node)

    def dictify(self, obj):
        """Extended to handle JSON properties."""

        dict_ = {}
        for node in self:

            name = node.name
            try:
                if is_index_property(obj, name):
                    value = getattr(obj, name)
                else:
                    getattr(self.inspector.column_attrs, name)
                    value = getattr(obj, name)

            except AttributeError:

                try:
                    # Classic colanderalchemy
                    prop = getattr(self.inspector.relationships, name)

                    # We know this node is good to pass through as is, don't try to dictify subitems
                    if isinstance(node.typ, ModelSchemaType):
                        value = getattr(obj, name)

                    elif prop.uselist:
                        value = [self[name].children[0].dictify(o)
                                 for o in getattr(obj, name)]
                    else:
                        o = getattr(obj, name)
                        value = None if o is None else self[name].dictify(o)
                except AttributeError:
                    # The given node isn't part of the SQLAlchemy model
                    msg = 'SQLAlchemySchemaNode.dictify: %s not found on %s'
                    logger.debug(msg, name, self)
                    continue

            # SQLAlchemy mostly converts values into Python types
            #  appropriate for appstructs, but not always.  The biggest
            #  problems are around `None` values so we're dealing with
            #  those here.  All types should accept `colander.null` so
            #  we mostly change `None` into that.

            if value is None:
                if isinstance(node.typ, colander.String):
                    # colander has an issue with `None` on a String type
                    #  where it translates it into "None".  Let's check
                    #  for that specific case and turn it into a
                    #  `colander.null`.
                    dict_[name] = colander.null
                else:
                    # A specific case this helps is with Integer where
                    #  `None` is an invalid value.  We call serialize()
                    #  to test if we have a value that will work later
                    #  for serialization and then allow it if it doesn't
                    #  raise an exception.  Hopefully this also catches
                    #  issues with user defined types and future issues.
                    try:
                        node.serialize(value)
                    except Exception:
                        dict_[name] = colander.null
                    else:
                        dict_[name] = value
            else:
                dict_[name] = value

        return dict_

    def objectify(self, dict_, context=None):
        """Extended to handle JSON properties."""

        mapper = self.inspector
        context = mapper.class_() if context is None else context

        # If our schema and widgets wants pass us back full objects instead of theri dictified versions, let them pass through
        if sqlalchemy.inspect(dict_, raiseerr=False) is not None:
            return dict_

        for attr in dict_:
            if mapper.has_property(attr):
                prop = mapper.get_property(attr)

                if hasattr(prop, 'mapper'):
                    value = dict_[attr]

                    if prop.uselist:
                        # Sequence of objects

                        if isinstance(value, ModelSetResultList):
                            # We know we do not need to try to convert these
                            pass
                        else:
                            # Try to map incoming colander items back to SQL items
                            value = [self[attr].children[0].objectify(obj)
                                     for obj in dict_[attr]]
                    else:

                        if hasattr(value, "__tablename__"):
                            # Raw SQLAlchemy object - do not try to convert
                            pass
                        else:
                            # Single object
                            if value:
                                value = self[attr].objectify(value)
                else:
                    value = dict_[attr]
                    if value is colander.null:
                        # `colander.null` is never an appropriate
                        # value to be placed on an SQLAlchemy object
                        # so we translate it into `None`.
                        value = None
                setattr(context, attr, value)
            elif hasattr(context, attr):
                # Set any properties on the object which are not SQLAlchemy column based.
                # These are JSONBProperty like user_data and password on the user model (actual column is called _password, but we mangle the password has before pushing it through)
                value = dict_[attr]
                setattr(context, attr, value)
            else:
                # Ignore attributes if they are not mapped
                logger.debug(
                    'SQLAlchemySchemaNode.objectify: %s not found on '
                    '%s. This property has been ignored.',
                    attr, self
                )
                continue

        return context

    def get_schema_from_column(self, prop, overrides):
        """Extended to handle JSON/JSONB.
        """

        # The name of the SchemaNode is the ColumnProperty key.
        name = prop.key
        kwargs = dict(name=name)
        column = prop.columns[0]

        typedecorator_overrides = getattr(column.type,
                                          self.ca_class_key, {}).copy()
        # print("colanderalchemy ", prop, typedecorator_overrides)

        declarative_overrides = column.info.get(self.sqla_info_key, {}).copy()
        self.declarative_overrides[name] = declarative_overrides.copy()

        key = 'exclude'

        if key not in itertools.chain(declarative_overrides, overrides) \
           and typedecorator_overrides.pop(key, False):
            log.debug('Column %s skipped due to TypeDecorator overrides', name)
            return None

        if key not in overrides and declarative_overrides.pop(key, False):
            log.debug('Column %s skipped due to declarative overrides', name)
            return None

        if overrides.pop(key, False):
            log.debug('Column %s skipped due to imperative overrides', name)
            return None

        self.check_overrides(name, 'name', typedecorator_overrides,
                             declarative_overrides, overrides)

        for key in ['missing', 'default']:
            self.check_overrides(name, key, typedecorator_overrides, {}, {})

        # The SchemaNode built using the ColumnProperty has no children.
        children = []

        # The type of the SchemaNode will be evaluated using the Column type.
        # User can overridden the default type via Column.info or
        # imperatively using overrides arg in SQLAlchemySchemaNode.__init__

        # Support sqlalchemy.types.TypeDecorator

        dialect = self.dbsession.bind.engine.dialect
        column_type = column.type.dialect_impl(dialect)

        imperative_type = overrides.pop('typ', None)
        declarative_type = declarative_overrides.pop('typ', None)
        typedecorator_type = typedecorator_overrides.pop('typ', None)

        if self.type_overrides is not None:

            type_overrides_type, type_overrides_kwargs = self.type_overrides(self, name, column, column_type)

            if type_overrides_type == TypeOverridesHandling.drop:
                # This column should not appear on the form
                log.debug('Column %s: dropped by type overrides callback', name)
                return None

            elif type_overrides_type == TypeOverridesHandling.unknown:
                # type overrides callback doesn't know about this column
                type_overrides_type = None
                type_overrides_kwargs = {}

        else:
            type_overrides_type = None
            type_overrides_kwargs = {}

        if imperative_type is not None:
            if hasattr(imperative_type, '__call__'):
                type_ = imperative_type()
            else:
                type_ = imperative_type
            log.debug('Column %s: type overridden imperatively: %s.', name, type_)

        elif declarative_type is not None:
            if hasattr(declarative_type, '__call__'):
                type_ = declarative_type()
            else:
                type_ = declarative_type
            log.debug('Column %s: type overridden via declarative: %s.',
                      name, type_)

        elif typedecorator_type is not None:
            if hasattr(typedecorator_type, '__call__'):
                type_ = typedecorator_type()
            else:
                type_ = typedecorator_type
            log.debug('Column %s: type overridden via TypeDecorator: %s.',
                      name, type_)

        elif type_overrides_type is not None:
            if hasattr(type_overrides_type, '__call__'):
                type_ = type_overrides_type()
            else:
                type_ = type_overrides_type
            log.debug('Column %s: type overridden via type_overrides: %s.',
                      name, type_)

        elif isinstance(column_type, Boolean):
            type_ = colander.Boolean()

        elif isinstance(column_type, Date):
            type_ = colander.Date()

        elif isinstance(column_type, DateTime):
            type_ = colander.DateTime(default_tzinfo=None)

        elif isinstance(column_type, Enum):
            type_ = colander.String()
            kwargs["validator"] = colander.OneOf(column.type.enums)

        elif isinstance(column_type, Float):
            type_ = colander.Float()

        elif isinstance(column_type, Integer):
            type_ = colander.Integer()

        elif isinstance(column_type, String):
            type_ = colander.String()
            kwargs["validator"] = colander.Length(0, column.type.length)

        elif isinstance(column_type, Numeric):
            type_ = colander.Decimal()

        elif isinstance(column_type, Time):
            type_ = colander.Time()
        else:
            raise NotImplementedError(
                'Not able to derive a colander type from sqlalchemy '
                'type: %s  Please explicitly provide a colander '
                '`typ` for the "%s" Column.'
                % (repr(column_type), name)
            )

        """
        Add default values

        possible values for default in SQLA:
         1. plain non-callable Python value
              - give to Colander as a default
         2. SQL expression (derived from ColumnElement)
              - leave default blank and allow SQLA to fill
         3. Python callable with 0 or 1 args
            1 arg version takes ExecutionContext
              - leave default blank and allow SQLA to fill

        all values for server_default should be ignored for
        Colander default
        """
        if (isinstance(column.default, ColumnDefault) and column.default.is_scalar):
            kwargs["default"] = column.default.arg

        """
        Add missing values

        possible values for default in SQLA:
         1. plain non-callable Python value
              - give to Colander as a missing unless nullable
         2. SQL expression (derived from ColumnElement)
              - set missing to 'drop' to allow SQLA to fill this in
                and make it an unrequired field
         3. Python callable with 0 or 1 args
            1 arg version takes ExecutionContext
              - set missing to 'drop' to allow SQLA to fill this in
                and make it an unrequired field

        if nullable, then missing = colander.null (this has to be
        the case since some colander types won't accept `None` as
        a value, but all accept `colander.null`)

        all values for server_default should result in 'drop'
        for Colander missing

        autoincrement results in drop
        """
        if isinstance(column.default, ColumnDefault):
            if column.default.is_callable:
                kwargs["missing"] = drop
            elif column.default.is_clause_element:  # SQL expression
                kwargs["missing"] = drop
            elif column.default.is_scalar:
                kwargs["missing"] = column.default.arg
        elif column.nullable:
            kwargs["missing"] = colander.null
        elif isinstance(column.server_default, FetchedValue):
            kwargs["missing"] = drop  # value generated by SQLA backend
        elif (hasattr(column.table, "_autoincrement_column") and id(column.table._autoincrement_column) == id(column)):
            # this column is the autoincrement column, so we can drop
            # it if it's missing and let the database generate it
            kwargs["missing"] = drop

        kwargs.update(type_overrides_kwargs)
        kwargs.update(typedecorator_overrides)
        kwargs.update(declarative_overrides)
        kwargs.update(overrides)

        return colander.SchemaNode(type_, *children, **kwargs)

    def get_schema_from_relationship(self, prop, overrides):
        """Extended for property support."""

        # The name of the SchemaNode is the ColumnProperty key.
        name = prop.key
        kwargs = dict(name=name)
        declarative_overrides = prop.info.get(self.sqla_info_key, {}).copy()
        self.declarative_overrides[name] = declarative_overrides.copy()

        class_ = prop.mapper

        if declarative_overrides.pop('exclude', False):
            log.debug('Relationship %s skipped due to declarative overrides',
                      name)
            return None

        for key in ['name', 'typ']:
            self.check_overrides(name, key, {}, declarative_overrides,
                                 overrides)

        key = 'children'
        imperative_children = overrides.pop(key, None)
        declarative_children = declarative_overrides.pop(key, None)
        if imperative_children is not None:
            children = imperative_children
            msg = 'Relationship %s: %s overridden imperatively.'
            log.debug(msg, name, key)

        elif declarative_children is not None:
            children = declarative_children
            msg = 'Relationship %s: %s overridden via declarative.'
            log.debug(msg, name, key)

        else:
            children = None

        key = 'includes'
        imperative_includes = overrides.pop(key, None)
        declarative_includes = declarative_overrides.pop(key, None)
        if imperative_includes is not None:
            includes = imperative_includes
            msg = 'Relationship %s: %s overridden imperatively.'
            log.debug(msg, name, key)

        elif declarative_includes is not None:
            includes = declarative_includes
            msg = 'Relationship %s: %s overridden via declarative.'
            log.debug(msg, name, key)

        else:
            includes = None

        # TODO: Utmost piece of garbage here.. remove this
        if not self.automatic_relationships:
            if name not in (self.includes or ()):
                log.debug("Does not construct relationship for %s unless explicitly included", name)
                return None

        if self.relationship_overrides:

            result = self.relationship_overrides(self, name, prop, class_)
            if result == TypeOverridesHandling.drop:
                return None
            elif result == TypeOverridesHandling.unknown:
                pass
            else:
                assert isinstance(result, colander.SchemaNode)
                return result

        key = 'excludes'
        imperative_excludes = overrides.pop(key, None)
        declarative_excludes = declarative_overrides.pop(key, None)

        if imperative_excludes is not None:
            excludes = imperative_excludes
            msg = 'Relationship %s: %s overridden imperatively.'
            log.debug(msg, name, key)

        elif declarative_excludes is not None:
            excludes = declarative_excludes
            msg = 'Relationship %s: %s overridden via declarative.'
            log.debug(msg, name, key)

        else:
            excludes = None

        # see issue #25 in ColanderAlchemy for possible patch
        if includes is None and excludes is None:
            includes = [p.key for p in inspect(class_).column_attrs]

        key = 'overrides'
        imperative_rel_overrides = overrides.pop(key, None)
        declarative_rel_overrides = declarative_overrides.pop(key, None)

        if imperative_rel_overrides is not None:
            rel_overrides = imperative_rel_overrides
            msg = 'Relationship %s: %s overridden imperatively.'
            log.debug(msg, name, key)

        elif declarative_rel_overrides is not None:
            rel_overrides = declarative_rel_overrides
            msg = 'Relationship %s: %s overridden via declarative.'
            log.debug(msg, name, key)

        else:
            rel_overrides = None

        # Add default values for missing parameters.
        if prop.innerjoin:
            # Inner joined relationships imply it is mandatory
            missing = required
        else:
            # Any other join is thus optional
            missing = []
        kwargs['missing'] = missing

        kwargs.update(declarative_overrides)
        kwargs.update(overrides)

        if children is not None:
            if prop.uselist:
                # xToMany relationships.
                return SchemaNode(Sequence(), *children, **kwargs)
            else:
                # xToOne relationships.
                return SchemaNode(Mapping(), *children, **kwargs)

        node = PropertyAwareSQLAlchemySchemaNode(class_,
                                    name=name,
                                    includes=includes,
                                    excludes=excludes,
                                    overrides=rel_overrides,
                                    missing=missing,
                                    type_overrides=self.type_overrides,
                                    relationship_overrides=self.relationship_overrides)

        if prop.uselist:
            node = SchemaNode(Sequence(), node, **kwargs)

        node.name = name

        return node

    def clone(self):
        cloned = self.__class__(self.class_,
                                self.dbsession,
                                self.includes,
                                self.excludes,
                                self.overrides,
                                self.unknown,
                                self.nested,
                                self.type_overrides,
                                self.relationship_overrides,
                                self.automatic_relationships,
                                **self.kwargs)
        cloned.__dict__.update(self.__dict__)
        cloned.children = [node.clone() for node in self.children]
        return cloned



"""Default user model field definitions.

This module defines what fields the default user implementation can have. You can subclass these mixins and then provide your own implementation for concrete models.
"""

# Standard Library
import datetime
from uuid import uuid4

# SQLAlchemy
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import inspection
from sqlalchemy.ext.indexable import index_property
from sqlalchemy.orm.session import Session

# Websauna
# from sqlalchemy.dialects.postgresql import JSONB
from websauna.system.model.columns import INET
from websauna.system.model.columns import JSONB
from websauna.system.model.columns import UUID
from websauna.system.model.columns import UTCDateTime
from websauna.system.model.json import NestedMutationDict
from websauna.utils.crypt import generate_random_string
from websauna.utils.time import now


#: Initialze user_data JSONB structure with these fields on new User
DEFAULT_USER_DATA = {
    "full_name": None,

    # The initial sign up method (email, phone no, imported, Facebook) for this user
    "registration_source": None,

    # Is it the first time this user is logging to our system? If it is then take the user to fill in the profile page.
    "first_login": True,

    "social": {
        # Each of the social media login data imported here as it goes through SocialLoginMapper.import_social_media_user()
    }
}


class UserMixin:
    """A user who signs up with email or with email from social media.

    This mixin provides the default required columns for user model in Websauna.

    The user contains normal columns and then ``user_data`` JSON field where properties and non-structured data can be easily added without migrations. This is especially handy to store incoming OAuth fields from social networks. Think Facebook login data and user details.
    """

    #: Running counter id of the user
    id = Column(Integer, autoincrement=True, primary_key=True)

    #: Publicly exposable ID of the user
    uuid = Column(UUID(as_uuid=True), default=uuid4)

    #: Though not displayed on the site, the concept of "username" is still preversed. If the site needs to have username (think Instragram, Twitter) the user is free to choose this username after the sign up. Username is null until the initial user activation is completed after db.flush() in create_activation().
    username = Column(String(256), nullable=True, unique=True)

    email = Column(String(256), nullable=True, unique=True)

    #: Stores the password + hash + cycles as password hasher internal format.. By default uses Argon 2 format. See :py:meth:`websauna.system.Initializer.configure_password`
    hashed_password = Column('password', String(256), nullable=True)

    #: When this account was created
    created_at = Column(UTCDateTime, default=now)

    #: When the account data was updated last time
    updated_at = Column(UTCDateTime, onupdate=now)

    #: When this user was activated: email confirmed or first social login
    activated_at = Column(UTCDateTime, nullable=True)

    #: Is this user account enabled. The support can disable the user account in the case of suspected malicious activity.
    enabled = Column(Boolean(name="user_enabled_binary"), default=True)

    #: When this user accessed the system last time. None if the user has never logged in (only activation email sent). Information stored for the security audits.
    last_login_at = Column(UTCDateTime, nullable=True)

    #: From which IP address did this user log in from. If this IP is null the user has never logged in (only activation email sent). Information stored for the security audits. It is also useful for identifying the source country of users e.g. for localized versions.
    last_login_ip = Column(INET, nullable=True)

    #: Misc. user data as a bag of JSON. Do not access directly, but use JSONBProperties below
    user_data = Column(NestedMutationDict.as_mutable(JSONB), default=DEFAULT_USER_DATA)

    #: Store when this user changed the password or authentication details. Updating this value causes the system to drop all sessions which were created before this moment. E.g. you will kick out all old sessions on a password or email change.
    last_auth_sensitive_operation_at = Column(UTCDateTime, nullable=True, default=now)

    #: Full name of the user (if given)
    full_name = index_property("user_data", "full_name")

    #: How this user signed up to the site. May include string like "email", "facebook" or "dummy". Up to the application to use this field. Default social media logins and email sign up set this.
    registration_source = index_property("user_data", "registration_source")

    #: Social media data of the user as a dict keyed by user media
    social = index_property("user_data", "social")

    #: Is this the first login the user manages to do to our system. If this flag is set the user has not logged in to the system before and you can give warm welcoming experience.
    first_login = index_property("user_data", "first_login")

    @property
    def friendly_name(self) -> str:
        """How we present the user's name to the user itself.

        Picks one of 1) full name if set 2) username if set 3) email.
        """
        full_name = self.full_name
        if full_name:
            return full_name

        # Get the username if it looks like non-automatic form
        if self.username:
            if self.username.startswith("user-"):
                return self.email
            else:
                return self.username

        return self.email

    def generate_username(self) -> str:
        """The default username we give for the user."""
        assert self.id
        return "user-{}".format(self.id)

    def is_activated(self) -> bool:
        """Has the user completed the email activation."""
        return self.activated_at is not None

    def can_login(self) -> bool:
        """Is this user allowed to login."""
        # TODO: is_active defined in Horus
        return self.enabled and self.is_activated

    def is_in_group(self, name) -> bool:

        # TODO: groups - defined in Horus
        for g in self.groups:
            if g.name == name:
                return True
        return False

    def is_admin(self) -> bool:
        """Does this user the see the main admin interface link.

        TODO: This is very suboptimal, wasted database cycles, etc. Change this.
        """
        return self.is_in_group(GroupMixin.DEFAULT_ADMIN_GROUP_NAME)

    def is_valid_session(self, session_created_at: datetime.datetime) -> bool:
        """Check if the current session is still valid for this user."""
        return self.last_auth_sensitive_operation_at <= session_created_at

    def __str__(self):
        return self.friendly_name

    def __repr__(self):
        return "#{}: {}".format(self.id, self.friendly_name)


class GroupMixin:
    """Basic fields for Websauna default group model."""

    #: Assign the first user initially to this group
    DEFAULT_ADMIN_GROUP_NAME = "admin"

    #: Running counter id of the group
    id = Column(Integer, autoincrement=True, primary_key=True)

    #: Publicly exposable ID of the group
    uuid = Column(UUID(as_uuid=True), default=uuid4)

    #: Human readable / machine referrable name of the group
    name = Column(String(64), unique=True)

    #: Human readable description of the group
    description = Column(String(256))

    #: When this group was created.
    created_at = Column(UTCDateTime, default=now)

    #: When the group was updated last time. Please note that this does not concern group membership, only desription updates.
    updated_at = Column(UTCDateTime, onupdate=now)

    #: Extra JSON data to be stored with this group
    group_data = Column(NestedMutationDict.as_mutable(JSONB), default=dict)


class SiteCreator:
    """Component responsible for setting up an empty site on first login.

    The site creator is run by the activation of the first user. This either happens¨

    * When the activation email is sent to the first user

    * When the first user logs through social media account

    """

    def init_empty_site(self, dbsession: Session, user: UserMixin):
        """When the first user signs up build the admin groups and make the user member of it.

        Make the first member of the site to be admin and superuser.
        """
        # Try to reflect related group class based on User model
        i = inspection.inspect(user.__class__)
        Group = i.relationships["groups"].mapper.entity

        # Do we already have any groups... if we do we probably don'¨t want to init again
        if dbsession.query(Group).count() > 0:
            return
        g = Group(name=Group.DEFAULT_ADMIN_GROUP_NAME)
        dbsession.add(g)
        g.users.append(user)

    def check_empty_site_init(self, dbsession: Session, user: UserMixin):
        """Call after user creation to see if this user is the first user and should get initial admin rights."""
        assert user.id, "Please flush your db"

        # Try to reflect related group class based on User model
        i = inspection.inspect(user.__class__)
        Group = i.relationships["groups"].mapper.entity

        # If we already have groups admin group must be there
        if dbsession.query(Group).count() > 0:
            return

        self.init_empty_site(dbsession, user)


class ActivationMixin:
    """Sign up / forgot password activation code reference."""

    #: Running counter id
    id = Column(Integer, autoincrement=True, primary_key=True)

    #: When this group was created.
    created_at = Column(UTCDateTime, default=now)

    #: When the group was updated last time. Please note that this does not concern group membership, only desription updates.
    updated_at = Column(UTCDateTime, onupdate=now)

    #: All activation tokens must have expiring time
    expires_at = Column(UTCDateTime, nullable=False)

    code = Column(String(32), nullable=False, unique=True, default=lambda: generate_random_string(32))

    def is_expired(self):
        """The activation best before is past and we should not use it anymore."""
        return self.expires_at < now()


class UserGroupMixin:
    """Map users to groups."""

    id = Column(Integer, autoincrement=True, primary_key=True)



"""SQLAlchemy form support."""
# Standard Library
import typing as t

# Pyramid
from pyramid.interfaces import IRequest

# SQLAlchemy
from sqlalchemy.orm import Query
from sqlalchemy.orm import Session

# Websauna
from websauna.system.http import Request

from . import CRUD as _CRUD
from . import Resource as _Resource


class Resource(_Resource):
    """Maps one SQLAlchemy model instance to a traversable URL path.

    Describe how to display SQLAlchemy objects in breadcrumbs.
    """

    def get_title(self):
        """Title on show / edit / delete pages."""
        return "{parent_title} #{id}".format(
            parent_title=self.__parent__.title,
            id=self.obj.id
        )


class CRUD(_CRUD):
    """SQLAlchemy CRUD controller.

    A traversing endpoint which maps listing, add, edit and delete views for an SQLAlchemy model.
    """

    def __init__(self, request: IRequest, model: t.Optional[type]=None):
        """Create a CRUD root resource for a given model.

        :param request: Current HTTP Request.
        :param model: Can be set on class level or instance level.
        """
        super(CRUD, self).__init__(request)

        if model is not None:
            self.model = model

    def get_model(self):
        """Get the SQLAlchemy model instance we are managing.

        :return: SQLAlchemy model instance.
        """
        return self.model

    def get_dbsession(self) -> Session:
        """Override to use a different database session.

        Default to ``request.dbsession``.

        :return: A database session.
        """
        return self.request.dbsession

    def get_query(self) -> Query:
        """Get SQLAlchemy Query object which we use to populate this listing.

        Views can specify their own queries - e.g. filter by user. This is just the default for everything.

        :return: SQLAlchemy query.
        """
        model = self.get_model()
        dbsession = self.get_dbsession()
        return dbsession.query(model)

    def delete_object(self, obj):
        """Delete one item in the CRUD.

        Called by delete view if no alternative logic is implemented.
        :param obj: Raw object, e.g. SQLAlchemy instance.
        """
        dbsession = self.get_dbsession()
        return dbsession.delete(obj)

    def fetch_object(self, id):
        """Pull a raw object from the database.

        Use the ``get_query()`` to get the query base and then return the object with matching id.

        First check for legal ids and raise KeyError to signal that the traversed ``id`` might be actually a view name.
        """
        model = self.get_model()

        column_name = self.mapper.mapping_attribute

        column_instance = getattr(model, column_name, None)
        assert column_instance, "Model {} does not define column/attribute {} used for CRUD resource traversing".format(self.model, column_name)

        obj = self.get_query().filter(column_instance == id).first()
        if not obj:
            raise KeyError("Object id {} was not found for CRUD {} using model {}".format(id, self, model))

        return obj


def sqlalchemy_deleter(view: object, context: Resource, request: Request):
    """A view callback to delete item in SQLAlchemy CRUD.

    :param view: View object.
    :param context: Traversal context
    :param request: Current HTTP Request.
    """
    obj = context.get_object()
    dbsession = request.dbsession
    dbsession.delete(obj)



"""Session management."""
# Standard Library
import functools
import logging
import os
import pickle as cPickle
import typing as t
from urllib.parse import urlparse

# Pyramid
from pyramid.request import Request
from pyramid.session import signed_serialize

from pyramid_redis_sessions import RedisSession
from pyramid_redis_sessions import _delete_cookie
from pyramid_redis_sessions import _generate_session_id
from pyramid_redis_sessions import _get_session_id_from_cookie
from pyramid_redis_sessions import get_default_connection
from pyramid_redis_sessions.util import _parse_settings
from pyramid_redis_sessions.util import get_unique_session_id
from pyramid_redis_sessions.util import persist

# Websauna
from websauna.utils.time import now


logger = logging.getLogger(__name__)


#: Todo temporary fix until we manage address this properly in pyramid_redis_sessions
NO_SESSION_FILE_EXTENSIONS = [".js", ".css", ".ico", ".png", ".gif", ".jpg"]


def ignore_session(url: str) -> bool:
    """Should we ignore session for this request?

    Static assets requests do not need a session, exception made to requests to /notebook, because those requests are proxies to another daemon.
    :param url: Request url.
    :return: Flag indicating if session should be ignored.
    """
    path = urlparse(url).path
    to_notebook = path.startswith('/notebook/')
    ext = str(os.path.splitext(path)[1]).lower()
    return True if (ext in NO_SESSION_FILE_EXTENSIONS and not to_notebook) else False


class WebsaunaSession(RedisSession):
    """A specialized session handler that supports initial parameters.

    We can pass `initial_data` that pre-populates session data keys when the session is written for the first time. Usually this is when CSRF token is generated.

    .. note::Move this to upstream pyramid_redis_session - its development has stalled for now.
    """

    def __init__(
        self,
        initial_data: dict,
        redis,
        session_id,
        new,
        new_session,
        serialize=cPickle.dumps,
        deserialize=cPickle.loads,
    ):
        super().__init__(redis, session_id, new, new_session, serialize, deserialize)
        self.initial_data = initial_data

    @persist
    def __setitem__(self, key, value):
        # Check if do not have any values in the session and then initialize its data
        if self.managed_dict:
            self.managed_dict.update(self.initial_data)
        self.managed_dict[key] = value

    def get_csrf_token(self):
        token = self.get('_csrft_', None)
        if token is None:
            token = self.new_csrf_token()
        else:
            token = str(token)
        return token


def _set_cookie(
    session,
    request,
    response,
    cookie_name,
    cookie_max_age,
    cookie_path,
    cookie_domain,
    cookie_secure,
    cookie_httponly,
    secret,
):
    """
    `session` is via functools.partial
    `request` and `response` are appended by add_response_callback
    """
    cookieval = signed_serialize(session.session_id, secret)
    response.set_cookie(
        cookie_name,
        value=cookieval,
        max_age=cookie_max_age,
        path=cookie_path,
        domain=cookie_domain,
        secure=cookie_secure,
        httponly=cookie_httponly,)


def _cookie_callback(
    session,
    request,
    response,
    session_cookie_was_valid,
    cookie_on_exception,
    set_cookie,
    delete_cookie,
    cookieless_headers,
):
    """
    Response callback to set the appropriate Set-Cookie header.
    `session` is via functools.partial
    `request` and `response` are appended by add_response_callback
    """

    # The view signals this is cacheable response
    # and we should not stamp a session cookie on it
    for header in cookieless_headers:
        if header in response.headers:
            return

    if session._invalidated:
        if session_cookie_was_valid:
            delete_cookie(response=response)
        return

    # Do not session cookie if we have not written anything to session yet
    # has_content = len(session.keys()) > 0
    # if session.new and has_content:
    if session.new:  # TODO: Wait until Pyramid 1.9 fixes land Websauna and then we can address this properly

        if cookie_on_exception is True or request.exception is None:

            set_cookie(request=request, response=response)

            # If we set a cookie we need to make sure that downstream
            # web serves do not serve this response from a cache
            # for requests coming in with a different session cookie.
            # Otherwise we might leak sessions between users.
            varies = ("Cookie",)
            vary = set(response.vary if response.vary is not None else [])
            vary |= set(varies)
            response.vary = vary

        elif session_cookie_was_valid:
            # We don't set a cookie for the new session here (as
            # cookie_on_exception is False and an exception was raised), but we
            # still need to delete the existing cookie for the session that the
            # request started with (as the session has now been invalidated).
            delete_cookie(response=response)


def WebsaunaSessionFactory(
    secret,
    timeout=1200,
    cookie_name='session',
    cookie_max_age=None,
    cookie_path='/',
    cookie_domain=None,
    cookie_secure=False,
    cookie_httponly=True,
    cookie_on_exception=True,
    url=None,
    host='localhost',
    port=6379,
    db=0,
    password=None,
    socket_timeout=None,
    connection_pool=None,
    encoding='utf-8',
    encoding_errors='strict',
    unix_socket_path=None,
    client_callable=None,
    serialize=cPickle.dumps,
    deserialize=cPickle.loads,
    id_generator=_generate_session_id,
    cookieless_headers=('expires', 'cache-control'),
    klass=WebsaunaSession,
):
    """
    Overrides the RedisSessionFactory with Websauna specifi functionality.

    .. note ::

        Due to functional paradigm, there was no clean way to override this.
        Move this to upstream pyramid_redis_session - its development has stalled for now

    Constructs and returns a session factory that will provide session data
    from a Redis server. The returned factory can be supplied as the
    ``session_factory`` argument of a :class:`pyramid.config.Configurator`
    constructor, or used as the ``session_factory`` argument of the
    :meth:`pyramid.config.Configurator.set_session_factory` method.

    Parameters:

    ``secret``
    A string which is used to sign the cookie.

    ``timeout``
    A number of seconds of inactivity before a session times out.

    ``cookie_name``
    The name of the cookie used for sessioning. Default: ``session``.

    ``cookie_max_age``
    The maximum age of the cookie used for sessioning (in seconds).
    Default: ``None`` (browser scope).

    ``cookie_path``
    The path used for the session cookie. Default: ``/``.

    ``cookie_domain``
    The domain used for the session cookie. Default: ``None`` (no domain).

    ``cookie_secure``
    The 'secure' flag of the session cookie. Default: ``False``.

    ``cookie_httponly``
    The 'httpOnly' flag of the session cookie. Default: ``True``.

    ``cookie_on_exception``
    If ``True``, set a session cookie even if an exception occurs
    while rendering a view. Default: ``True``.

    ``url``
    A connection string for a Redis server, in the format:
    redis://username:password@localhost:6379/0
    Default: ``None``.

    ``host``
    A string representing the IP of your Redis server. Default: ``localhost``.

    ``port``
    An integer representing the port of your Redis server. Default: ``6379``.

    ``db``
    An integer to select a specific database on your Redis server.
    Default: ``0``

    ``password``
    A string password to connect to your Redis server/database if
    required. Default: ``None``.

    ``client_callable``
    A python callable that accepts a Pyramid `request` and Redis config options
    and returns a Redis client such as redis-py's `StrictRedis`.
    Default: ``None``.

    ``serialize``
    A function to serialize the session dict for storage in Redis.
    Default: ``cPickle.dumps``.

    ``deserialize``
    A function to deserialize the stored session data in Redis.
    Default: ``cPickle.loads``.

    ``id_generator``
    A function to create a unique ID to be used as the session key when a
    session is first created.
    Default: private function that uses sha1 with the time and random elements
    to create a 40 character unique ID.

    ``cookieless_headers``
    If view has set any of these response headers do not add a session
    cookie on this response. This way views generating cacheable content,
    like images, can signal the downstream web server that this content
    is safe. Otherwise if we set a cookie on these responses it could
    result to user session leakage.

    The following arguments are also passed straight to the ``StrictRedis``
    constructor and allow you to further configure the Redis client::

      socket_timeout
      connection_pool
      encoding
      encoding_errors
      unix_socket_path
    """
    def factory(request, initial_data, new_session_id=get_unique_session_id):
        redis_options = dict(
            host=host,
            port=port,
            db=db,
            password=password,
            socket_timeout=socket_timeout,
            connection_pool=connection_pool,
            encoding=encoding,
            encoding_errors=encoding_errors,
            unix_socket_path=unix_socket_path,
        )

        # an explicit client callable gets priority over the default
        redis = client_callable(request, **redis_options) \
            if client_callable is not None \
            else get_default_connection(request, url=url, **redis_options)

        # attempt to retrieve a session_id from the cookie
        session_id_from_cookie = _get_session_id_from_cookie(
            request=request,
            cookie_name=cookie_name,
            secret=secret,
        )

        new_session = functools.partial(
            new_session_id,
            redis=redis,
            timeout=timeout,
            serialize=serialize,
            generator=id_generator,
        )

        if session_id_from_cookie and redis.exists(session_id_from_cookie):
            session_id = session_id_from_cookie
            session_cookie_was_valid = True
        else:
            session_id = new_session()
            session_cookie_was_valid = False

        session = klass(
            initial_data,
            redis=redis,
            session_id=session_id,
            new=not session_cookie_was_valid,
            new_session=new_session,
            serialize=serialize,
            deserialize=deserialize,
        )

        set_cookie = functools.partial(
            _set_cookie,
            session,
            cookie_name=cookie_name,
            cookie_max_age=cookie_max_age,
            cookie_path=cookie_path,
            cookie_domain=cookie_domain,
            cookie_secure=cookie_secure,
            cookie_httponly=cookie_httponly,
            secret=secret,
        )
        delete_cookie = functools.partial(
            _delete_cookie,
            cookie_name=cookie_name,
            cookie_path=cookie_path,
            cookie_domain=cookie_domain,
        )
        cookie_callback = functools.partial(
            _cookie_callback,
            session,
            session_cookie_was_valid=session_cookie_was_valid,
            cookie_on_exception=cookie_on_exception,
            set_cookie=set_cookie,
            delete_cookie=delete_cookie,
            cookieless_headers=cookieless_headers,
        )
        request.add_response_callback(cookie_callback)

        return session

    return factory


def set_creation_time_aware_session_factory(config):
    """Setup a session factory that remembers time when the session was created.

    We need this information to later invalidate session for the authentication change details.
    """
    settings = config.registry.settings

    # special rule for converting dotted python paths to callables
    for option in ('client_callable', 'serialize', 'deserialize', 'id_generator'):
        key = 'redis.sessions.{option}'.format(option=option)
        if key in settings:
            settings[key] = config.maybe_dotted(settings[key])

    options = _parse_settings(settings)
    session_factory = WebsaunaSessionFactory(**options)

    def create_session(request: Request) -> t.Union[WebsaunaSession, t.Dict]:
        """Given a request, return a WebsaunaSession or an empty dict.

        :param request: A client request.
        :return: An instance of WebsaunaSession or an empty dict
        """
        # Pass in the the data we use to track session on the server side.
        # Esp. created_at is used to later manually purge old sessions
        initial_data = {
            "client_addr": request.client_addr,
            "created_at": now(),
        }

        # Make sure we do not get accidental race conditions when populating sessions when browser asynchronously fetches more resources
        url = request.url
        if ignore_session(url):
            logger.debug("Skipped session creation for {url}".format(url=url))
            return {}

        session = session_factory(request, initial_data)

        # Allow session backend to use this information to debug session problems
        session.url = url
        return session

    config.set_session_factory(create_session)



def test_sample_html_email(web_server, browser):
    """Test that HTML email renderer works."""
    browser.visit(web_server + "/sample-html-email")
    assert browser.is_element_present_by_css("#heading-sample-email")



"""A test URL you can call to cause a production run time error."""
# Standard Library
import logging

# Pyramid
from pyramid.view import view_config

# Websauna
from websauna.system.core.loggingcapture import get_logging_user_context


logger = logging.getLogger(__name__)


@view_config(route_name='error_trigger')
def error_trigger(request):
    """An error logging view to generate a run-time error.
    """
    logger.debug("Logging debug message on debug level")
    logger.info("Logging debug message on info level")
    logger.warn("Logging debug message on warning level")
    user_context = get_logging_user_context(request)
    logger.error("Logging debug message on error level", exc_info=True, extra={"user": user_context})
    logger.fatal("Logging debug message on fatal level")

    raise RuntimeError("Test error.")



"""ws-pserve script.

DEPRECATED: Wrapper around Pyramid pserve script.
"""
# Standard Library
import sys
import typing as t

# Websauna
from websauna.system.devop.scripts import proxy_to_pyramid_script


try:
    import gevent.monkey
    gevent.monkey.patch_all()
except ImportError:
    pass


def main(argv: t.List[str]=sys.argv):
    """Proxy to Pyramid pserve script.

    This script is deprecated and will be removed in Websauna 1.0.0
    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    proxy_to_pyramid_script('pserve', argv)



"""Mutable JSON data support for SQLAlchemy.

Based on the original Kotti CMS implementation:

* https://github.com/Kotti/Kotti/blob/5a33384e7b11994371c415b489edbec88ecbf044/kotti/sqla.py#L79
"""
# Standard Library
import json

# SQLAlchemy
from sqlalchemy import event
from sqlalchemy import types
from sqlalchemy.dialects.postgresql.json import JSON
from sqlalchemy.dialects.postgresql.json import JSONB
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm import mapper
from sqlalchemy.sql.schema import Column
from sqlalchemy_utils.types.json import JSONType


def _default(obj):
    if isinstance(obj, MutationDict):
        return obj._d
    elif isinstance(obj, MutationList):
        return obj._d


def json_serializer(d):
    """MutationDict friendly json_serializer for create_engine().

    See :py:meth:`websauna.system.model.meta.get_engine`.

    http://stackoverflow.com/a/36438671/315168
    """
    return json.dumps(d, default=_default)


class WebsaunaFriendlyMutable(Mutable):
    """A patched mutable that can deal with our generic column types."""

    @classmethod
    def as_mutable(cls, orig_sqltype):
        """Mark the value as nested mutable value.

        What happens here

        * We coerce the return value - the type value set on sqlalchemy.Column() to the underlying SQL typ

        * We mark this type value with a marker attribute

        * Then we set a global SQAlchemy mapper event handler

        * When mapper is done setting up our model classes, it will call the event handler for all models

        * We check if any of the models columns have our marked type value as the value

        * If so we call ``associate_with_attribute`` for this model and column that sets up ``MutableBase._listen_on_attribute`` event handlers. These event handlers take care of taking the raw dict coming out from database and wrapping it to NestedMutableDict.

        :param orig_sqltype: Usually websauna.system.model.column.JSONB instance
        :return: Marked and coerced type value
        """

        # Create an instance of this type and add a marker attribute,
        # so we later find it.
        # We cannot directly compare the result type values, as looks like
        # the type value might be mangled by dialect specific implementations
        # or lost somewhere. Never figured this out 100%.
        sqltype = types.to_instance(orig_sqltype)
        sqltype._column_value_id = id(sqltype)

        def listen_for_type(mapper, class_):
            for prop in mapper.column_attrs:
                # The original implementation has SQLAlchemy type comparator.
                # Here we need to be little more complex, because we define a type alias
                # for generic JSONB implementation
                if getattr(prop.columns[0].type, "_column_value_id", None) == sqltype._column_value_id:
                    cls.associate_with_attribute(getattr(class_, prop.key))

        event.listen(mapper, 'mapper_configured', listen_for_type)

        return sqltype


class MutationDict(WebsaunaFriendlyMutable):
    """http://www.sqlalchemy.org/docs/orm/extensions/mutable.html
    """
    _wraps = dict

    def __init__(self, data):
        self._d = data
        super(MutationDict, self).__init__()

    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutationDict):
            if isinstance(value, dict):
                return cls(value)
            return Mutable.coerce(key, value)
        else:
            return value

    def __json__(self, request=None):
        return dict([(key, value.__json__(request))
                     if hasattr(value, '__json__') else (key, value)
                     for key, value in self._d.items()])


class MutationList(WebsaunaFriendlyMutable):
    _wraps = list

    def __init__(self, data):
        self._d = data
        super(MutationList, self).__init__()

    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutationList):
            if isinstance(value, list):
                return cls(value)
            return Mutable.coerce(key, value)
        else:
            return value

    def __radd__(self, other):
        return other + self._d

    def __json__(self, request=None):
        return [item.__json__(request) if hasattr(item, '__json__') else item
                for item in self._d]


def _make_mutable_method_wrapper(wrapper_class, methodname, mutates):
    def replacer(self, *args, **kwargs):
        method = getattr(self._d, methodname)
        value = method(*args, **kwargs)
        if mutates:
            self.changed()
        return value

    replacer.__name__ = methodname
    return replacer


for wrapper_class in (MutationDict, MutationList):
    for methodname, mutates in (
            ('__iter__', False),
            ('__len__', False),
            ('__eq__', False),
            ('__add__', False),
            ('__getitem__', False),
            ('__getslice__', False),
            ('__repr__', False),
            ('get', False),
            ('keys', False),
            ('values', False),
            ('items', False),

            ('__setitem__', True),
            ('__delitem__', True),
            ('__setslice__', True),
            ('__delslice__', True),
            ('append', True),
            ('clear', True),
            ('extend', True),
            ('insert', True),
            ('pop', True),
            ('setdefault', True),
            ('update', True),
            ('remove', True),
    ):
        # Only wrap methods that do exist on the wrapped type!
        if getattr(wrapper_class._wraps, methodname, False):
            setattr(
                wrapper_class, methodname,
                _make_mutable_method_wrapper(
                    wrapper_class, methodname, mutates),
            )


class NestedMixin(object):
    """Base class to to nested dict and list state tracking."""

    #: Pointer to parent NestedMutatedDict or NestedMutatedList. If the parent is Column then this is None.
    __parent__ = None

    def __init__(self, *args, **kwargs):
        self.__parent__ = kwargs.pop('__parent__', None)
        super(NestedMixin, self).__init__(*args, **kwargs)

    def __getitem__(self, key):
        value = self._d.__getitem__(key)
        return self.try_wrap(value)

    def changed(self):

        if self.__parent__ is not None:
            # Direct dict or tuple parent here
            self.__parent__.changed()
        else:
            # Parent is SQLALchemy model instance, let Mutable base class take over
            super(NestedMixin, self).changed()

    def try_wrap(self, value):
        for typ, wrapper in MUTATION_WRAPPERS.items():
            if isinstance(value, typ):
                value = wrapper(value, __parent__=self)
                break
        return value

    def __eq__(self, other):
        return self._d == other

    def __str__(self):
        return "#{} {}: {}".format(id(self), self.__class__, self._d)

    def __repr__(self):
        return self.__str__()


class NestedMutationDict(NestedMixin, MutationDict):
    def setdefault(self, key, default):
        if isinstance(default, list):
            default = NestedMutationList(default, __parent__=self)
        elif isinstance(default, dict):
            default = NestedMutationDict(default, __parent__=self)
        return super(NestedMutationDict, self).setdefault(key, default)


class NestedMutationList(NestedMixin, MutationList):
    pass


MUTATION_WRAPPERS = {
    dict: NestedMutationDict,
    list: NestedMutationList,
}


def wrap_as_nested(key: str, data: object, parent: object) -> object:
    """Convert raw dict or list data to a NestedMutable version with a parent pointer set.

    After this is in-place, the object dirty state should be automatically tracked.

    :param key: Column name
    :param data: dict or list
    :param parent: SQLAlchemy model instance
    :return: data in the nested mutable wrapper
    """

    for typ, wrapper in MUTATION_WRAPPERS.items():
        if isinstance(data, typ):

            # We do data.copy() here to ensure we don't run into "Mutable defaults" Python issue
            # __parent__ = None means we dont' have parent dict or list
            data = wrapper(data.copy(), __parent__=None)  # TODO: do a deep copy?

            # SQLAlchemy model instance is directly parent of this NestedMutate
            data._parents[parent] = key
            return data

    # Unknown type :(
    return data


def is_json_like_column(c: Column) -> bool:
    """Check if the colum."""
    return isinstance(c.type, (JSONType, JSON, JSONB))


def _get_column_default(target: object, column: Column):
    if column.default is not None:
        if callable(column.default.arg):
            return column.default.arg(target)
        else:
            return column.default.arg


def setup_default_value_handling(cls):
    """A SQLAlchemy model class decorator that ensures JSON/JSONB default values are correctly handled.

    Settings default values for JSON fields have a bunch of issues, because dict and list instances need to be wrapped in ``NestedMutationDict`` and ``NestedMutationList`` that correclty set the parent object dirty state if the container content is modified. This class decorator sets up hooks, so that default values are automatically converted to their wrapped versions.

    .. note ::

        We are only concerned about initial values. When values are loaded from the database, ``Mutable`` base class of ``NestedMutationDict`` correctly sets up the parent pointers.

    .. note ::

        This must be applid to the class directly, as SQLAlchemy does not seem ot pick it up if it's applied to an (abstract) parent class.

    This might be addressed in future SQLAlchemy / Websauna versions so that we can get rid of this ugly little decorator.
    """

    @event.listens_for(cls, "init")
    def init(target, args, kwargs):
        for c in target.__table__.columns:
            if is_json_like_column(c):
                default = _get_column_default(target, c)
                default = wrap_as_nested(c.name, default, target)
                setattr(target, c.name, default)

    return cls


def init_for_json(cls):
    """Check if we need to add JSON column specific SQLAlchemy event listers for this model."""
    # import pdb ; pdb.set_trace()
    global _loaded_listener
    has_json_columns = any([is_json_like_column(c) for c in cls.__table__.columns])
    if has_json_columns:
        setup_default_value_handling(cls)



"""pyramid_layout panel rendering helper.

TODO: This module is on the list of those who will get burial by fire.
"""
# Pyramid
from pyramid_layout.interfaces import IPanel
from zope.interface import providedBy

from markupsafe import Markup


# Jinja 2 adoption of render_panel
# TODO: make this configurable in upstream
def render_panel(context, request, name='', *args, **kw):
    """Render a panel.

    TODO: Clean this up

    Renders the named panel, returning a `unicode` object that is the
    rendered HTML for the panel.  The panel is looked up using the current
    context (or the context given as keyword argument, to override the
    context in which the panel is called) and an optional given name
    (which defaults to an empty string).
    The panel is called passing in the current
    context, request and any additional parameters passed into the
    `render_panel` call.  In case a panel isn't found, `None` is returned.
    """
    if "controls" not in kw:
        kw["controls"] = True

    adapters = request.registry.adapters
    panel = adapters.lookup((providedBy(context),), IPanel, name=name)
    assert panel, "Could not find panel {name}".format(name=name)
    return Markup(panel(context, request, *args, **kw))



"""Export settings and secrets as environment variables to subprocesses."""
# Standard Library
import os

# Pyramid
from pyramid.registry import Registry

# SQLAlchemy
from sqlalchemy.engine.url import make_url

# Websauna
from websauna.system.core.utils import get_secrets


def create_settings_env(registry: Registry):
    """Create os.environ where have exported all settings and secrets.

    This is used for subprocess to create a child processes which are aware of our settings.  All dots are replaced with underscores. The exported environment varible is either prefixed with ``main``  or ``secret``. The exported environment variables are uppercased. Parent process environment variables are automatically inherited.

    The environment will look like::

        MAIN_websauna_SITE_ID=foo
        SECRETS_AWS_CONSUMER_KEY=baa

    You will also have::

        MAIN_SQL_HOST
        MAIN_SQL_DATABASE
        MAIN_SQL_USER
        MAIN_SQL_PASSWORD

    Integers, booleans and objects are exported as strings:

        MAIN_SOME_BOOLEAN=True
        MAIN_SOME_INTEGER=123

    """
    env = os.environ.copy()

    settings = registry.settings
    secrets = get_secrets(registry)

    # Export database credentials
    url = make_url(settings["sqlalchemy.url"])

    env["MAIN_SQL_HOST"] = url.host or ""
    env["MAIN_SQL_DATABASE"] = url.database or ""
    env["MAIN_SQL_USERNAME"] = url.username or ""
    env["MAIN_SQL_PASSWORD"] = url.password or ""

    for key, val in settings.items():
        key = "main.{}".format(key)
        key = key.replace(".", "_")
        env[key.upper()] = str(val)

    for key, val in secrets.items():
        key = "secret.{k}".format(k=key)
        key = key.replace(".", "_")
        env[key.upper()] = str(val)

    return env



# Pyramid
from pyramid.authentication import \
    SessionAuthenticationPolicy as _SessionAuthenticationPolicy

# Websauna
from websauna.utils.time import now


class SessionAuthenticationPolicy(_SessionAuthenticationPolicy):
    """Session authentication policy which makes sure all responses get vary: Cookie.

    Originally from https://github.com/pypa/warehouse/blob/master/warehouse/accounts/auth_policy.py
    """

    #: Session has this key set to the timestamp when the user authentication happened. The key is removed when the user logs out.
    authenticated_at_key = "authenticated_at"

    #: This is the timestamp when this session logged out last time
    unauthenticated_at_key = "unauthenticated_at"

    def remember(self, request, userid, **kw):
        """ Store a userid in the session."""
        request.session[self.userid_key] = userid

        # Do not lose the original log in timestamp if we get multiple calls to remember()
        if self.authenticated_at_key not in request.session:
            request.session[self.authenticated_at_key] = now()

        return []

    def forget(self, request):
        """User logs out or is forced to be forgotten."""

        # Wiggle session keys so that we know when unauthentication happened
        request.session[self.unauthenticated_at_key] = now()

        # Remove authenticated_at timestamp so nothing can read it when there is no logged in associated user
        if self.authenticated_at_key in request.session:
            del request.session[self.authenticated_at_key]

        return super().forget(request)

    def unauthenticated_userid(self, request):
        """What is the user id for unauthenticated users."""

        # If we're calling into this API on a request, then we want to register
        # a callback which will ensure that the response varies based on the
        # Cookie header.
        # request.add_response_callback(add_vary_callback_if_cookie("Cookie"))

        # Dispatch to the real SessionAuthenticationPolicy
        return super().unauthenticated_userid(request)



# Pyramid
from pyramid.security import Allow
from pyramid.security import Authenticated
from pyramid.security import Everyone
from zope.interface import implementer

# Websauna
from websauna.system.core.interfaces import IRoot
from websauna.system.core.traversal import Resource


@implementer(IRoot)
class Root(Resource):
    """Pyramid routing root with default permission set up.

    These permission mappings are used unless you supply your own traversing context. For the sake of simplicity, we only declare one permission named ``authenticated`` which is given to all authenticated users.

    Permissions are as:

    * authenticated: Pseudopermission given to all authenticated users

    * superuser: equal to have SSH access to the website - can execute arbitrary Python code

    For more information see :ref:`permissions`.
    """

    __acl__ = [

        # Creaet pseudo-permission "authenticated" for all authenticated users
        (Allow, Authenticated, 'authenticated'),

        # See Notebook
        (Allow, "superuser:superuser", 'shell'),

        # All traversable resources are accesible to public by default. This permission is primarily used by sitemap to construct publicly traversable site hierarchy.
        (Allow, Everyone, 'view'),
    ]

    def __init__(self, request):
        self.request = request
        self.__name__ = ""

    @classmethod
    def root_factory(cls, request):
        """Called by Pyramid routing framework to create a new root for a request."""
        return cls(request)

    def get_title(self):
        """Title used in breadcrumbs."""
        return "Home"



"""The user authentication helper functions."""
# System
import typing as t

# Websauna
from websauna.system.http import Request
from websauna.system.user.models import User
from websauna.system.user.utils import get_user_registry


def get_user(session_token: str, request: Request) -> t.Optional[User]:
    """Extract the logged in user from the request object using Pyramid's authentication framework."""
    # user_id = unauthenticated_userid(request)
    # TODO: Abstract this to its own service like in Warehouse?
    user_registry = get_user_registry(request)
    user = None
    if session_token is not None:
        user = user_registry.get_user_by_session_token(session_token)
        # Check through conditions why this user would no longer be valid
        if user and not user.can_login():
            # User account disabled while in mid-session
            user = None
    return user


def get_request_user(request: Request) -> t.Optional[User]:
    """Reify method for request.user"""
    user_id = request.unauthenticated_userid
    return get_user(user_id, request) if user_id else None
